
/*  ----------------------------------------------------------------------
   sf Firewall Software -- a TCP/IP packet filter for Linux
   Copyright (C) 1997 Robert Muchsel and Roland Schmid 

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

/*
 * sf Firewall Relay - Communication with Configuration Server and Client
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef OS2
#include "sf_custom.h"
#include "sf_custom_client.h"
#ifdef SOLARIS
#define sys_siglist _sys_siglist
#define NSIG _sys_nsig
int gethostname(char *, int);
char *tempnam(const char *, const char *);
char *strdup(const char *);
#endif
#else /* OS2 */
#include <io.h>
#include "sf_custom_os2.h"
#endif

#include "sf_relay.h"

/*
 * Public variables
 */

/*
 * Static variables and other local stuff
 */

static int guiservsock;         /* server for connection to client */
static int guiconnsock;         /* connected client connection */
static int guiconnected = 0;    /* connection to client established */
static struct sockaddr_in guiservaddr; /* server address */
static struct sockaddr_in guicliaddr;  /* client address */
static struct in_addr localaddr[MAX_LOCAL_ADDR]; /* local addresses */
static int localaddrno = 0;     /* number of local addresses */

static int sfconnected = 0;     /* number of established sf connections */
static int sfsocks[MAX_SERVERS];/* sockets for connections to sf daemon */
static int sfstate[MAX_SERVERS];/* state of connections to sf daemon */
static struct sockaddr_in sfaddr[MAX_SERVERS]; /* remote addresses */
static char *sfbuf[MAX_SERVERS];/* dynamic buffers */
static int sfbufsize[MAX_SERVERS]; /* size of dynamic buffers */
static FILE *sffds[MAX_SERVERS];/* file descriptors for RECEIVE_FILE */

static FILE *fdlog = NULL;      /* file pointer to log file */
static time_t ltime = 0;        /* the current time */
static char lstime[20] = "";    /* the current time in ASCII */
static char buffer[BUFFER_SIZE];/* read buffer */
static char clibuf[CLIBUF_SIZE];/* read buffer for GUI client connection */
static int clibufoffset = 0;    /* number of valid bytes in clibuf */
static int write_success = 0;   /* last write file successful */
static FILE *clifd;		/* file descriptor for WRITE_LOCAL_FILE */
static int clifd_valid = 0;     /* true if file is open for writing */

static char *tmpfiles[MAX_TEMPFILES]; /* array to store temp filenames */
static int tmpfileno = 0;             /* number of currently existing temp files */

#ifdef RELAY_DEBUG
static void debug_entry(int i, char *msg, int status)
{ char stmp[20];                /* string for IP address to string conversion */

  strcpy(stmp, inet_ntoa(sfaddr[i].sin_addr));
  if (status)
    fprintf(fdlog, "%s DEBUG: Server conn. %i to %s:%hu\n"
                   "                        %s\n"
                   "                        conn_state[%i] == %x\n",
            lstime, i, stmp, ntohs(sfaddr[i].sin_port),
            msg,
            i, sfstate[i]);
  else
    fprintf(fdlog, "%s DEBUG: Client conn. %i to %s:%hu\n"
                   "                        %s\n",
            lstime, i, stmp, ntohs(sfaddr[i].sin_port), msg);
}
#endif

/*
 * Get the current time, convert it to ASCII and store both
 * values in static variables.
 */

static void get_time(void)
{
  /* Get current time */
  time(&ltime);

  /* Convert the current time to a string. Don't use ctime(), because
     the ctime() output is too verbose to be useful. */

  strftime(lstime, sizeof(lstime), TIME_FORMAT, localtime(&ltime));
}

/*
 * Remove temporary files
 */

static void remove_tempfiles()
{ int i;

  for (i=0; i < tmpfileno; i++) {
    unlink(tmpfiles[i]);
    free(tmpfiles[i]);
  }
  tmpfileno = 0;
}

/*
 * Add temporary filename to array
 */

static void add_tempfile(char *name)
{
  if (tmpfileno >= MAX_TEMPFILES) 
    remove_tempfiles();

  tmpfiles[tmpfileno++] = name;
}
  
/*
 * Clean up before terminating
 */

static void cleanup(int sig)
{
  get_time();

  fprintf(fdlog, "%s relay daemon terminated by signal %i: %s\n",
          lstime, sig,
#ifndef OS2
          sys_siglist[sig]
#else
          ""
#endif
          );
  fclose(fdlog);
  remove_tempfiles();

  exit(0);
}

/*
 * Initialize variables, set up socket ...
 */

static void initialize()
{ int i;
  char hname[256];
  struct hostent *hp;

  /* use TCP connection for DNS queries */
  sethostent(1);

  /* open log file */
  if ((fdlog = fopen(RELAY_LOG, "a")) == NULL) {
    perror("Cannot append to log file");
    exit(1);
  }

  /* install signal handlers */
  for (i = 1; i < NSIG; i++)
    signal(i, cleanup);

  /* signals to ignore */
  signal(SIGHUP, SIG_IGN);
#ifndef OS2
  signal(SIGURG, SIG_IGN);
  signal(SIGCONT, SIG_IGN);
  signal(SIGIO, SIG_IGN);
#endif
  signal(SIGPIPE, SIG_IGN);     /* test for EPIPE after each write() */

  /* The log file must not be fully buffered... We want to be able to read
   * it even if the system crashes.
   */
  setvbuf(fdlog, NULL, _IOLBF, 0);

  for (i=0; i<MAX_SERVERS; i++) {
    sfstate[i]=STATE_NOCONN;
    sfbuf[i]=NULL;
    sfbufsize[i]=0;
  }

  /* determine local IP addresses */
  if (gethostname(hname, sizeof(hname)) != 0) {
    perror("Cannot get local host name");
    fprintf(fdlog, "%s Cannot get local host name\n", lstime);
    exit(1);
  }
  if ((hp = gethostbyname(hname)) == NULL) {
    perror("Cannot resolve local host name");
    fprintf(fdlog, "%s Cannot resolve local host name\n", lstime);
    exit(1);
  }
  for (i=0; (i < MAX_LOCAL_ADDR) && (hp->h_addr_list[i] != NULL); i++)
    memcpy(&(localaddr[i]), hp->h_addr_list[i], sizeof(struct in_addr));
  localaddrno = i;


  if ((guiservsock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("Cannot create socket");
    fprintf(fdlog, "%s Cannot create socket\n", lstime);
    exit(1);
  }

  i = 1;
  if (setsockopt(guiservsock, SOL_SOCKET, SO_REUSEADDR,
                 (void *) &i, sizeof(i)) < 0) {
    perror("WARNING: Cannot set SO_REUSEADDR option on socket");
    fprintf(fdlog, "%s WARNING: Cannot set SO_REUSEADDR option on socket\n", lstime);
  }

  memset(&guiservaddr, 0, sizeof(guiservaddr));
  guiservaddr.sin_family = AF_INET;
  guiservaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  guiservaddr.sin_port = htons(RELAY_PORT);

  if (bind(guiservsock, (struct sockaddr *) &guiservaddr, sizeof(guiservaddr)) < 0) {
    perror("Cannot bind address to socket");
    fprintf(fdlog, "%s Cannot bind address to socket\n", lstime);
    exit(1);
  }

  if (listen(guiservsock, RELAY_BACKLOG) < 0) {
    perror("Cannot listen on socket");
    fprintf(fdlog, "%s Cannot listen on socket\n", lstime);
    exit(1);
  }

  fprintf(fdlog, "%s relay daemon started\n", lstime);

}

/*
 * Close all server connections
 */

static void close_all_conns()
{ int i;

  for (i=0; i<MAX_SERVERS; i++)
    if (sfstate[i]) {
      if ((sfstate[i] & STATE_WAITDATA) == STATE_WAITDATA)
        free(sfbuf[i]);
      sfbufsize[i]=0;
      sfstate[i]=STATE_NOCONN;
      close(sfsocks[i]);
    }
  sfconnected = 0;
}

/*
 * Read data from socket, returns number of bytes read, negative value on error
 */

#define READ_ECLOSE     -1      /* connection has been closed by client */
#define READ_EFATAL     -2      /* any read error that should not occur,
                                   will be handled like aborted connection */

static int read_data(int s) /* s: socket descriptor */
{
  int retval;

  retval = read(s, buffer, sizeof(buffer));

  if (retval < 0) {

    /* read terminated by signal -> return 0 bytes read */
    if (errno == EINTR)
      return 0;

    if (errno == EPIPE)
      return READ_ECLOSE;
      /* According to the documentation only write() should
         return EPIPE, but read() does, too! So we handle
         EPIPE like EOF. */

    return READ_EFATAL;
  }

  if (retval == 0) /* EOF, i.e. client has closed the connection */
    return READ_ECLOSE;

  /* read successful: assert(retval>0) */

  return retval;

}

/*
 * Write data to socket
 */

#define WRITE_EPIPE     -1      /* write error EPIPE */
#define WRITE_EFATAL    -2      /* unexpected write error, connection closed */

static int write_data(int s, void *buf, size_t count) /* s: socket descriptor */
{
  void *bf = buf;
  size_t cnt = count;
  int retval = 0;
  char stmp[20];

  while (cnt > 0) {

    if ((retval = write(s, bf, cnt)) <= 0) {

      if (errno == EINTR)
        continue;               /* write interrupted by signal, retry */

      /* handle connection abort */
      if (s == guiconnsock) {
        strcpy(stmp, inet_ntoa(guicliaddr.sin_addr));

        if (errno == EPIPE)
          fprintf(fdlog, "%s GUI client conn. aborted by client: %s:%hu\n",
                  lstime, stmp, ntohs(guicliaddr.sin_port));
        else
          fprintf(fdlog, "%s GUI client conn. aborted (write error): %s:%hu\n",
                  lstime, stmp, ntohs(guicliaddr.sin_port));

        guiconnected --;
        close(guiconnsock);
        close_all_conns();
        remove_tempfiles();
        clibufoffset = 0;
      }
      else {
        int i;

        for (i=0; (i<MAX_SERVERS) && ((!sfstate[i]) || (sfsocks[i]!=s)); i++);
        if (i<MAX_SERVERS) { /* assert( (sfstate[i]) && (sfsocks[i]==s) ) */
          strcpy(stmp, inet_ntoa(sfaddr[i].sin_addr));

          if (errno == EPIPE)
            fprintf(fdlog, "%s Server connection aborted by server: %s:%hu\n",
                    lstime, stmp, ntohs(sfaddr[i].sin_port));
          else
            fprintf(fdlog, "%s Server connection aborted (write error): %s:%hu\n",
                    lstime, stmp, ntohs(sfaddr[i].sin_port));

          sfconnected--;
          if ((sfstate[i] & STATE_WAITDATA) == STATE_WAITDATA)
            free(sfbuf[i]);
          sfbufsize[i]=0;
          sfstate[i]=STATE_NOCONN;
          close(sfsocks[i]);
        }
      }

      if (errno == EPIPE)       /* client has closed connection */
        return WRITE_EPIPE;
      else                      /* handle any other error like connection termination */
        return WRITE_EFATAL;
    }

    cnt -= retval;
    bf = (void *) ((unsigned long) bf + (unsigned long) retval);
  }

  return count;
}

/*
 * Handle connection to sf daemon
 */

static void handle_server(int num)
{ int retval;
  char stmp[20];
  struct serv_data *sd;
  struct serv_data sdr;
  unsigned long sdmagic;
  unsigned long sdcommand;
  unsigned long sdlength;
#ifdef RELAY_DEBUG
  char tmp[255];
#endif

  if ((retval = read_data(sfsocks[num])) <= 0) {

    strcpy(stmp, inet_ntoa(sfaddr[num].sin_addr));

    if (retval == READ_ECLOSE)
      fprintf(fdlog, "%s Server connection aborted by server: %s:%hu\n",
              lstime, stmp, ntohs(sfaddr[num].sin_port));
    if (retval == READ_EFATAL)
      fprintf(fdlog, "%s Server connection aborted (read error): %s:%hu\n",
              lstime, stmp, ntohs(sfaddr[num].sin_port));

    close(sfsocks[num]);
    if ((sfstate[num] & STATE_WAITDATA) == STATE_WAITDATA)
      free(sfbuf[num]);
    sfbufsize[num]=0;
    sfstate[num] = STATE_NOCONN;
    sfconnected--;

    return;
  }

  /* read successful */
#ifdef RELAY_DEBUG
  sprintf(tmp, "read successful, retval == %i", retval);
  debug_entry(num, tmp, 1);
#endif

  restart:

  if ((sfstate[num] & STATE_WAITDATA) != STATE_WAITDATA) {

    if (retval < sizeof(struct serv_data)) { /* wait for more data */
      sfstate[num] |= STATE_WAITDATA;
      if ((sfbuf[num] = (char *)malloc(retval)) == NULL) {
        fprintf(fdlog, "%s Cannot allocate memory\n", lstime);
        exit(1);
      }
      sfbufsize[num]=retval;
      memcpy(sfbuf[num], buffer, retval);
      return;
    }

    sd = (struct serv_data *) buffer;
    sdmagic = ntohl(sd->magic);
    sdcommand = ntohl(sd->command);
    sdlength = ntohl(sd->length);

    if ((sdmagic != SERV_MAGIC) || (sdlength > SERV_MAXLEN)) { /* ignore data */
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Wrong magic or length > SERV_MAXLEN\n");
#endif
      return;
    }

    if ((retval - sizeof(struct serv_data)) < sdlength) { /* wait for more data */
      sfstate[num] |= STATE_WAITDATA;
      if ((sfbuf[num] = (char *)malloc(retval)) == NULL) {
        fprintf(fdlog, "%s Cannot allocate memory\n", lstime);
        exit(1);
      }
      sfbufsize[num] = retval;
      memcpy(sfbuf[num], buffer, retval);
      return;
    }
  }
  else { /* assert((sfstate[num] & STATE_WAITDATA) == STATE_WAITDATA) */
    int oldsize = sfbufsize[num];
    if ((sfbuf[num] = (char *) realloc(sfbuf[num], oldsize + retval)) == NULL) {
        fprintf(fdlog, "%s Cannot allocate memory\n", lstime);
        exit(1);
      }
    sfbufsize[num] += retval;
    sd = (struct serv_data *) sfbuf[num];
    sdmagic = ntohl(sd->magic);
    sdcommand = ntohl(sd->command);
    sdlength = ntohl(sd->length);

    memcpy(&(sfbuf[num][oldsize]), buffer, retval);

    if ((oldsize + retval) < sizeof(struct serv_data)) /* wait for more data */
      return;

    if ((sdmagic != SERV_MAGIC) || (sdlength > SERV_MAXLEN)) { /* ignore data */
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Wrong magic or length > SERV_MAXLEN\n");
#endif
      goto endproc;
    }

    if ((oldsize + retval - sizeof(struct serv_data)) < sdlength) /* wait for more data */
      return;
  }

  /* process data */
#ifdef RELAY_DEBUG
  sprintf(tmp, "processing command %lu, length %lu", sdcommand, sdlength);
  debug_entry(num, tmp, 1);
#endif

  sdr.magic = htonl(SERV_MAGIC);
  sdr.length = 0;

  switch (sdcommand) {
    case SERV_SERVER_HELLO:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_SERVER_HELLO\n");
#endif
      sdr.command = htonl(SERV_RELAY_HELLO);
      if (write_data(sfsocks[num], (void *) &sdr, sizeof(sdr)) < 0)
        return;
      sfstate[num] &= ~STATE_WAITHELLO;
      break;

    case SERV_RELAY_HELLO:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_RELAY_HELLO\n");
#endif
      break; /* ignore */

    case SERV_CLOSE:
      sdr.command = htonl(SERV_CLOSE_ACK);
      if (write_data(sfsocks[num], (void *) &sdr, sizeof(sdr)) >= 0) {
        strcpy(stmp, inet_ntoa(sfaddr[num].sin_addr));
        fprintf(fdlog, "%s Server connection closed by server: %s:%hu\n",
                lstime, stmp, ntohs(sfaddr[num].sin_port));

        close(sfsocks[num]);
        if ((sfstate[num] & STATE_WAITDATA) == STATE_WAITDATA)
          free(sfbuf[num]);
        sfbufsize[num] = 0;
        sfstate[num] = STATE_NOCONN;
        sfconnected--;
        return;
      }
      else /* write_data failed */
        return;
      break;

    case SERV_CLOSE_ACK:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_CLOSE_ACK\n");
#endif
      if ((sfstate[num] & STATE_CLOSEWAIT) != STATE_CLOSEWAIT)
        break; /* ignore CLOSE_ACK */
      strcpy(stmp, inet_ntoa(sfaddr[num].sin_addr));
      fprintf(fdlog, "%s Server connection closed by client: %s:%hu\n",
              lstime, stmp, ntohs(sfaddr[num].sin_port));

      close(sfsocks[num]);
      if ((sfstate[num] & STATE_WAITDATA) == STATE_WAITDATA)
        free(sfbuf[num]);
      sfbufsize[num] = 0;
      sfstate[num] = STATE_NOCONN;
      sfconnected--;
      return;

    case SERV_PING:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_PING\n");
#endif
      sdr.command = htonl(SERV_PING_REPLY);
      if (write_data(sfsocks[num], (void *) &sdr, sizeof(sdr)) < 0)
        return;
      break;

    case SERV_PING_REPLY :
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_PING_REPLY\n");
#endif
      sfstate[num] &= ~STATE_WAITPING;
      break;

    case SERV_FILE_DATA :
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_FILE_DATA\n");
#endif

      if ((sfstate[num] & STATE_RECEIVE_TMPFILE) == STATE_RECEIVE_TMPFILE) {
        if (sdlength == 0) { /* eof */
          sfstate[num] &= ~STATE_RECEIVE_TMPFILE;
          fclose(sffds[num]);
          sdr.magic = htonl(CLI_MAGIC);
          sdr.command = htonl(CLI_OK);
          sdr.length = 0;
          (void)write_data(guiconnsock, (void *) &sdr, sizeof(sdr));
          break;
        }
        if (fwrite(sd->data, 1, sdlength, sffds[num]) < sdlength)
          fprintf(fdlog, "%s Error writing to local file.\n", lstime);
      }
      else if ((sfstate[num] & STATE_RECEIVE_FILE) == STATE_RECEIVE_FILE) {
        if (sdlength == 0) /* eof */
          sfstate[num] &= ~STATE_RECEIVE_FILE;
        /* pass data to gui client */
        sd->magic = htonl(CLI_MAGIC);
        sd->command = htonl(CLI_FILE_DATA);
        /* leave length and data unchanged */
        if (write_data(guiconnsock, sd, sizeof(struct serv_data) + sdlength) < 0) {
          /* ignore rest of data */
          sfstate[num] &= ~STATE_RECEIVE_FILE;
          break;
        }
      }

      break;

    case SERV_WRITE_OK:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_WRITE_OK\n");
#endif
      write_success = 1;
      break;

    case SERV_WRITE_FAIL:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_WRITE_FAIL\n");
#endif
      write_success = 0;
      break;

    case SERV_OK :
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_OK\n");
#endif
      sdr.magic = htonl(CLI_MAGIC);
      sdr.command = htonl(CLI_OK);
      sdr.length = 0;
      (void)write_data(guiconnsock, (void *) &sdr, sizeof(sdr));
      break;

    case SERV_ERROR :
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received SERV_ERROR\n");
#endif
      sdr.magic = htonl(CLI_MAGIC);
      sdr.command = htonl(CLI_ERROR);
      sdr.length = 0;
      (void)write_data(guiconnsock, (void *) &sdr, sizeof(sdr));
      break;

    default:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Received unknown command: %lu\n", sdcommand);
#endif
      break;
  }

  endproc:

  /* free memory */
  if ((sfstate[num] & STATE_WAITDATA) == STATE_WAITDATA) {
    if ((sfbufsize[num] - sizeof(struct serv_data)) > sdlength) {
      /* beginning of next command may be in this packet */
      retval = sfbufsize[num] - sizeof(struct serv_data) - sdlength;
      memcpy(buffer, &(sfbuf[num][sizeof(struct serv_data) + sdlength]),
             retval);
      free(sfbuf[num]);
      sfbufsize[num] = 0;
      sfstate[num] &= ~STATE_WAITDATA;
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Data left, restarting function\n");
#endif
      goto restart;
    }
    free(sfbuf[num]);
    sfbufsize[num] = 0;
    sfstate[num] &= ~STATE_WAITDATA;
  }
  else
    if ((retval - sizeof(struct serv_data)) > sdlength) {
      /* beginning of next command may be in this packet */
      retval = retval - sizeof(struct serv_data) - sdlength;
      memmove(buffer, &(buffer[sizeof(struct serv_data) + sdlength]), retval);
#ifdef RELAY_DEBUG
      fprintf(fdlog, "                        Data left, restarting function\n");
#endif
      goto restart;
    }

  return;

}

/*
 * Open connection to sf daemon
 */
#define OPEN_EUNDEF             -1      /* unexpected error */
#define OPEN_EMAXCONN           -2      /* to many server connections */
#define OPEN_ECONNECT           -3      /* connect failed */
#define OPEN_ENEG               -4      /* initial negotiation failed */

static int open_server(struct in_addr addr)
{ int i;
  char stmp[20];

  strcpy(stmp, inet_ntoa(addr));

  for (i=0; (i<MAX_SERVERS) && (sfstate[i]); i++);
  if (i==MAX_SERVERS) {
    fprintf(fdlog, "%s Cannot connect to %s:%hu (too many connections)\n",
            lstime, stmp, SERVER_PORT);
    return OPEN_EMAXCONN;
  }

  if ((sfsocks[i] = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    fprintf(fdlog, "%s Cannot create socket for server connection.\n", lstime);
    return OPEN_EUNDEF;
  }

  memset(&(sfaddr[i]), 0, sizeof(sfaddr[i]));
  sfaddr[i].sin_family = AF_INET;
  sfaddr[i].sin_addr   = addr;
  sfaddr[i].sin_port   = htons(SERVER_PORT);

  if (connect(sfsocks[i], (struct sockaddr *) &(sfaddr[i]), sizeof(sfaddr[i])) < 0) {
    fprintf(fdlog, "%s Cannot connect to server %s:%hu\n",
            lstime, stmp, SERVER_PORT);
    close(sfsocks[i]);
    return OPEN_ECONNECT;
  }

  sfconnected++;
  sfstate[i] = STATE_CONN | STATE_WAITHELLO;
  fprintf(fdlog, "%s Connected to server %s:%hu\n", lstime, stmp, SERVER_PORT);

  while ( (sfstate[i]) && ((sfstate[i] & STATE_WAITHELLO) == STATE_WAITHELLO) )
    handle_server(i);

  if (!(sfstate[i]))
    return OPEN_ENEG;

  return i;
}


/*
 * find server connection
 */

static int find_conn(unsigned long *addr)
{ struct in_addr saddr;
  int i;

  saddr.s_addr = *addr;
#ifdef RELAY_DEBUG
  fprintf(fdlog, "%s address %s\n", lstime, inet_ntoa(saddr));
#endif

  for (i=0; (i<MAX_SERVERS) && (sfaddr[i].sin_addr.s_addr != saddr.s_addr); i++);
  if ((i<MAX_SERVERS) && (sfstate[i]))
    return i;
  else
    return -1;
}

/*
 * send file to client
 */

static void send_file(FILE *ftmp) 
{ char lbuf[CLIBUF_SIZE]; /* use local buffer */
  struct serv_data *cd = (struct serv_data *)lbuf;
  int buflen = CLIBUF_SIZE - sizeof(struct serv_data);
  int retval;

  cd->magic = htonl(CLI_MAGIC);
  cd->command = htonl(CLI_FILE_DATA);

  do {
    retval = fread(cd->data, 1, buflen, ftmp);
    cd->length = htonl(retval);
    if (write_data(guiconnsock, cd, sizeof(struct serv_data) + retval) < 0) {
      fclose(ftmp);
      return;
    }
  }
  while (retval > 0);

  fclose(ftmp);
}

/*
 * reverse lookup IP address
 */

struct hash {
  struct hash *next;
  int addr;
  char *name;
};

#define HASH_SIZE 199
struct hash *hashtable[HASH_SIZE];
#define HASH_KEY(x) (((unsigned) x) % HASH_SIZE)

void getname(int addr, char **name, int quiet)
{
  struct hostent *h_ent;
  unsigned key = HASH_KEY(addr);
  struct hash *h = hashtable[key];

  if (h == NULL)
    h = hashtable[key] = malloc(sizeof(struct hash));
  else {
    while ((h->addr != addr) && (h->next != NULL))
      h = h->next;

    if (h->addr == addr) {
      if (h->name != NULL)
        *name = h->name; 
      return;
    }

    h->next = malloc(sizeof(struct hash));
    h = h->next;
  }
  if (h == NULL) {
    fprintf(stderr, "malloc failed\n"); 
    return;
  }

  h->next = NULL;
  h->name = NULL;
  h->addr = addr;

  if (!quiet) {
    fprintf(stderr, "\r[%s]", *name); fflush(stderr);
  }
  h_ent = gethostbyaddr((char *) &addr, sizeof(struct in_addr), AF_INET);
  if (!quiet) {
    fprintf(stderr, "\r                 \r"); fflush(stderr);
  }
  if ((h_ent != NULL) && (h_ent->h_name != NULL) && (h_ent->h_name[0])) {
    h->name = strdup(h_ent->h_name);
    *name = h->name;
  }
}

/*
 * Handle connection to GUI client
 */

static void handle_client()
{ int retval;
  int i, j;
  char stmp[20];
  struct in_addr saddr;
  struct serv_data *sd = (struct serv_data *) buffer;
  struct serv_data *cd;
  struct serv_data cdr;
  unsigned long cdmagic;
  unsigned long cdcommand;
  unsigned long cdlength;
  char *tname;
  struct hostent *hp;
  struct servent *sp;
  FILE *ftmp;

#ifdef RELAY_DEBUG
  fprintf(fdlog, "%s handle_client()\n",lstime);
#endif

  if (clibufoffset >= sizeof(clibuf)) /* buffer full, this may not happen */
    clibufoffset = 0; /* throw away data */

  if ((retval = read(guiconnsock, &(clibuf[clibufoffset]),
                     sizeof(clibuf)-clibufoffset)) <= 0) {

    strcpy(stmp, inet_ntoa(guicliaddr.sin_addr));

    if (retval == 0)
      fprintf(fdlog, "%s GUI client conn. aborted by client: %s:%hu\n",
              lstime, stmp, ntohs(guicliaddr.sin_port));
    if (retval == -1)
      fprintf(fdlog, "%s GUI client conn. aborted (read error): %s:%hu\n",
              lstime, stmp, ntohs(guicliaddr.sin_port));

    close(guiconnsock);
    close_all_conns();
    remove_tempfiles();
    guiconnected--;

    return;
  }

  /* read successful */
  clibufoffset += retval;
#ifdef RELAY_DEBUG
  fprintf(fdlog, "%s read %i bytes from client, clibufoffset = %i\n",
          lstime, retval, clibufoffset);
#endif

  restart:

  if (clibufoffset < sizeof(struct serv_data))
    return; /* wait for more data */

  cd = (struct serv_data *) clibuf;
  cdmagic = ntohl(cd->magic);
  cdcommand = ntohl(cd->command);
  cdlength = ntohl(cd->length);

  if ((cdmagic != CLI_MAGIC) || (cdlength > CLI_MAXLEN)) { /* ignore data */
    clibufoffset = 0;
    return;
  }

  if ((clibufoffset - sizeof(struct serv_data)) < cdlength) /* wait for more data */
    return;
#ifdef RELAY_DEBUG
  fprintf(fdlog, "%s magic %lx, command %lx, length %li\n", lstime, cdmagic, cdcommand, cdlength);
#endif

  /* process data */
  sd->magic = htonl(SERV_MAGIC);
  sd->length = 0;
  cdr.magic = htonl(CLI_MAGIC);
  cdr.length = 0;
  switch (cdcommand) {

    case CLI_OPEN_SERV:
      if ((i = find_conn((unsigned long *)(cd->data))) >= 0) {
        cdr.command = htonl(CLI_ERROR_ISCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      saddr.s_addr = *((unsigned long *)(cd->data));
      if ((i = open_server(saddr)) < 0) {
        cdr.command = htonl(CLI_ERROR_OPENCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      break;

    case CLI_CLOSE_SERV:
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_CLOSE);
      if (write_data(sfsocks[i], sd, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sfstate[i] |= STATE_CLOSEWAIT;
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      break;

    case CLI_GET_FILE:
    case CLI_GET_FILE_TMP:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_GET_FILE%s\n",
              lstime, cdcommand == CLI_GET_FILE ? "" : "_TMP");
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      /* filename */
      strncpy(sd->data, &(cd->data[4]), BUFFER_SIZE-sizeof(struct serv_data));
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s CLI_GET_FILE(_TMP) filename %s|\n", lstime, sd->data);
#endif
      sd->length = htonl(strlen(sd->data) + 1);
      sd->command = htonl(SERV_GET_FILE);
      if (write_data(sfsocks[i], buffer,
                     sizeof(struct serv_data) + strlen(sd->data) + 1) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }

      if (cdcommand == CLI_GET_FILE_TMP) {
        tname = tempnam("/tmp", "sfr");
        add_tempfile(tname);
        if ((sffds[i] = fopen(tname, "w")) == NULL) {
          cdr.command = htonl(CLI_ERROR_FILE);
          if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
            return;
          break;
        }
        /* !!! permissions ? (applet must be allowed to read the file) */
        chmod(tname, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

        sd->magic = htonl(CLI_MAGIC);
        sd->command = htonl(CLI_FILE_TMP_NAME);
        sd->length = htonl(strlen(tname)+1);
        strcpy(sd->data, tname);
        if (write_data(guiconnsock, sd,
                       sizeof(struct serv_data) + strlen(sd->data) + 1) < 0)
          return;
        sfstate[i] |= STATE_RECEIVE_TMPFILE;
      }
      else {
        cdr.command = htonl(CLI_OK);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        sfstate[i] |= STATE_RECEIVE_FILE;
      }

      break;

    case CLI_WRITE_FILE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_WRITE_FILE %s\n", lstime, &(cd->data[4]));
#endif
      write_success = 0;
      for (i=0; i<MAX_SERVERS; i++)
        sfstate[i] &= ~STATE_WRITE_FILE;
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      /* filename */
      strncpy(sd->data, &(cd->data[4]), BUFFER_SIZE-sizeof(struct serv_data));
      sd->length = htonl(strlen(sd->data) + 1);
      sd->command = htonl(SERV_WRITE_FILE);
      if (write_data(sfsocks[i], buffer,
                     sizeof(struct serv_data) + strlen(sd->data) + 1) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      write_success = -1; /* waiting for status info from server */
      sfstate[i] |= STATE_WRITE_FILE;
      break;

    case CLI_FILE_DATA:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_FILE_DATA\n", lstime);
#endif
      if (clifd_valid) { // data for local file
        if (cdlength == 0) { /* eof */
          clifd_valid = 0;
          fclose(clifd);
        }
        else 
          if (fwrite(cd->data, 1, cdlength, clifd) < cdlength) {
            fprintf(fdlog, "%s Error writing to local file.\n", lstime);
            cdr.command = htonl(CLI_ERROR_FILE);
            if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
              return;
            break;
          }

        cdr.command = htonl(CLI_OK);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }

      for (i=0; (i<MAX_SERVERS) &&
                ((sfstate[i] & STATE_WRITE_FILE) != STATE_WRITE_FILE); i++);
      if (i == MAX_SERVERS) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cd->magic = htonl(SERV_MAGIC);
      cd->command = htonl(SERV_FILE_DATA);
      if (write_data(sfsocks[i], cd, sizeof(struct serv_data)+cdlength) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      break;

    case CLI_WRITE_STATUS:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_WRITE_STATUS, status %i\n", lstime, write_success);
#endif
      if (write_success == 1)
        cdr.command = htonl(CLI_OK);
      else if (write_success == 0)
        cdr.command = htonl(CLI_ERROR);
      else
        cdr.command = htonl(CLI_WAIT);
      write_success = 0;
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      break;

    case CLI_DELETE_FILE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_DELETE_FILE %s\n", lstime, cd->data);
#endif
      if (unlink(cd->data) == 0)
        cdr.command = htonl(CLI_OK);
      else
        cdr.command = htonl(CLI_ERROR_FILE);

      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;

      break;

    case CLI_RESOLVE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RESOLVE %s\n", lstime, cd->data);
#endif
      if (((hp = gethostbyname(cd->data)) == NULL) ||
          (hp->h_addr_list[0] == NULL)) {
        cdr.command = htonl(CLI_RESOLVE_REPLY);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->magic = htonl(CLI_MAGIC);
      sd->command = htonl(CLI_RESOLVE_REPLY);
      sd->length = htonl(4);
      memcpy(sd->data, hp->h_addr_list[0], 4);
      if (write_data(guiconnsock, sd, sizeof(struct serv_data) + 4) < 0)
        return;
      break;

    case CLI_RESOLVE_IP:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RESOLVE_IP\n", lstime);
#endif
      tname = NULL;
      getname(*((int *)cd->data), &tname, 1);
      if (tname == NULL || strlen(tname) == 0) {
        cdr.command = htonl(CLI_RESOLVE_REPLY);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->magic = htonl(CLI_MAGIC);
      sd->command = htonl(CLI_RESOLVE_REPLY);
      sd->length = htonl(strlen(tname)+1);
      memcpy(sd->data, tname, strlen(tname)+1);
      if (write_data(guiconnsock, sd, sizeof(struct serv_data) + 
        strlen(tname)+1) < 0)
        return;
      break;

    case CLI_RESOLVE_SERVICE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RESOLVE_SERVICE %s %s\n", lstime, cd->data, &(cd->data[strlen(cd->data)+1]) );
#endif
      if ((sp = getservbyname(cd->data, 
                   &(cd->data[strlen(cd->data)+1]))) == NULL) {
        cdr.command = htonl(CLI_RESOLVE_REPLY);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->magic = htonl(CLI_MAGIC);
      sd->command = htonl(CLI_RESOLVE_REPLY);
      sd->length = htonl(sizeof(int));
      j = htonl(ntohs((unsigned short)sp->s_port));
      memcpy(sd->data, &j, sizeof(int));
      if (write_data(guiconnsock, sd, sizeof(struct serv_data) + sizeof(int)) < 0)
        return;
      break;

    case CLI_RESOLVE_PORT:
      j = htons(ntohl(*((int *)cd->data)));
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RESOLVE_PORT %hu %s\n", lstime, ntohs(j), &(cd->data[4]));
#endif
      if (((sp = getservbyport(j, &(cd->data[4]))) == NULL) ||
          (sp->s_name == NULL)) {
        cdr.command = htonl(CLI_RESOLVE_REPLY);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->magic = htonl(CLI_MAGIC);
      sd->command = htonl(CLI_RESOLVE_REPLY);
      sd->length = htonl(strlen(sp->s_name)+1);
      memcpy(sd->data, sp->s_name, strlen(sp->s_name)+1);
      if (write_data(guiconnsock, sd, sizeof(struct serv_data) + 
        strlen(sp->s_name)+1) < 0)
        return;
      break;

    case CLI_READ_ADDR:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_READ_ADDR\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_READ_ADDR);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      sfstate[i] |= STATE_RECEIVE_FILE;

      break;

    case CLI_READ_RULES:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_READ_RULES\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_READ_RULES);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      sfstate[i] |= STATE_RECEIVE_FILE;

      break;

    case CLI_READ_VARS:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_READ_VARS\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_READ_VARS);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      sfstate[i] |= STATE_RECEIVE_FILE;

      break;

    case CLI_GET_TEMPNAME:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_GET_TEMPNAME\n", lstime);
#endif
      if ((tname = tempnam("/tmp", "sfr")) == NULL) {
        cdr.command = htonl(CLI_ERROR_FILE);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      add_tempfile(tname);
      sd->magic = htonl(CLI_MAGIC);
      sd->command = htonl(CLI_FILE_TMP_NAME);
      sd->length = htonl(strlen(tname)+1);
      strcpy(sd->data, tname);
      if (write_data(guiconnsock, sd,
                     sizeof(struct serv_data) + strlen(sd->data) + 1) < 0)
        return;
      break;

    case CLI_GET_LOCAL_FILE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_GET_LOCAL_FILE\n", lstime);
#endif
      if ((ftmp = fopen(cd->data, "r")) == NULL) {
        cdr.command = htonl(CLI_ERROR_FILE);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      send_file(ftmp);
      break;

    case CLI_WRITE_LOCAL_FILE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_WRITE_LOCAL_FILE\n", lstime);
#endif
      if ((clifd = fopen(cd->data, "w")) == NULL) {
        cdr.command = htonl(CLI_ERROR_FILE);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      /* !!! permissions ? (applet must be allowed to read the file) */
      chmod(cd->data, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      clifd_valid = 1;

      break;

    case CLI_CHECKCONFIG:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_CHECKCONFIG\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_CHECKCONFIG);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;

      break;

    case CLI_RECONFIG:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RECONFIG\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_RECONFIG);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;

      break;

    case CLI_READ_TCP:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_READ_TCP\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_READ_TCP);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;
      sfstate[i] |= STATE_RECEIVE_FILE;

      break;

    case CLI_KILL_TCP:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_KILL_TCP\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      memcpy(sd->data, &(cd->data[4]), cdlength-4);
      sd->length = htonl(cdlength-4);
      sd->command = htonl(SERV_KILL_TCP);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)+cdlength-4) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;

      break;

    case CLI_RESET_USAGE:
#ifdef RELAY_DEBUG
      fprintf(fdlog, "%s received CLI_RESET_USAGE\n", lstime);
#endif
      if ((i = find_conn((unsigned long *)(cd->data))) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      sd->command = htonl(SERV_RESET_USAGE);
      if (write_data(sfsocks[i], buffer, sizeof(struct serv_data)) < 0) {
        cdr.command = htonl(CLI_ERROR_NOTCONN);
        if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
          return;
        break;
      }
      cdr.command = htonl(CLI_OK);
      if (write_data(guiconnsock, &cdr, sizeof(cdr)) < 0)
        return;

      break;

    default:
      break;
  }

  clibufoffset = clibufoffset - sizeof(struct serv_data) - cdlength;

  if (clibufoffset > 0) { /* data left */
    memmove(clibuf, &(clibuf[sizeof(struct serv_data)+cdlength]), clibufoffset);
    goto restart;
  }

  clibufoffset = 0;

}

#ifdef RELAY_DEBUG
/*
 * Test function
 */
#define ST 2 /* sleep time between tests */

void testconn(int num)
{ struct serv_data sd;
  struct serv_data *sdp1, *sdp2;

  sd.magic  = htonl(SERV_MAGIC);
  sd.length = 0;

  /* send client hello first */
  sd.command = htonl(SERV_RELAY_HELLO);
  if (write_data(sfsocks[num], (void *) &sd, sizeof(sd)) < 0)
    return;
  sleep(ST);

  /* illegal magic */
  sd.magic = htonl(0xabc);
  if (write_data(sfsocks[num], (void *) &sd, sizeof(sd)) < 0)
    return;
  sleep(ST);

  /* length > MAX */
  sd.magic = htonl(SERV_MAGIC);
  sd.length = htonl(6000);
  if (write_data(sfsocks[num], (void *) &sd, sizeof(sd)) < 0)
    return;
  sleep(ST);

  /* split packet */
  sdp1 = (struct serv_data *)buffer;
  sdp1->magic = htonl(SERV_MAGIC);
  sdp1->length = 0;
  sdp1->command = htonl(SERV_PING_REPLY);
  if (write_data(sfsocks[num], (void *) buffer, 2) < 0)
    return;
  sleep(1);
  if (write_data(sfsocks[num], (void *) &(buffer[2]), 10) < 0)
    return;
  sleep(ST);

  /* split packet 2 */
  sdp1->length = htonl(100);
  sdp1->command = htonl(99);
  if (write_data(sfsocks[num], (void *) buffer, 32) < 0)
    return;
  sleep(1);
  if (write_data(sfsocks[num], (void *) &(buffer[32]), 80) < 0)
    return;
  sleep(ST);

  /* wrong length */
  if (write_data(sfsocks[num], (void *) buffer, 150) < 0)
    return;
  sleep(ST);

  /* two packets */
  sdp2 = (struct serv_data *) &(buffer[112]);
  sdp2->magic = htonl(SERV_MAGIC);
  sdp2->length = htonl(10);
  sdp2->command = htonl(98);
  if (write_data(sfsocks[num], (void *) buffer, 134) < 0)
    return;
  sleep(ST);

  /* two packets, first split */
  if (write_data(sfsocks[num], (void *) buffer, 50) < 0)
    return;
  sleep(1);
  if (write_data(sfsocks[num], (void *) &(buffer[50]), 84) < 0)
    return;
  sleep(ST);

  /* send big packet */
  memset(buffer, 65, BUFFER_SIZE);
  sdp1->magic = htonl(SERV_MAGIC);
  sdp1->length = htonl(BUFFER_SIZE - sizeof(struct serv_data));
  sdp1->command = htonl(33);
  if (write_data(sfsocks[num], buffer, BUFFER_SIZE) < 0)
    return;
  sleep(ST);

  /* send close */
  sdp1->length = 0;
  sdp1->command = htonl(SERV_CLOSE);
  sfstate[num] |= STATE_CLOSEWAIT;
  if (write_data(sfsocks[num], buffer, 12) < 0)
    return;

}
#endif

/*
 * Main loop, this function does not return
 */

static void main_loop()
{ fd_set rfds; /* bitmap for select call (reads) */
  int selret;  /* return value from select call */
  int maxfd;   /* highest used file descriptor, needed for select */
  int tempsock;
  struct sockaddr_in tempaddr;
  int tempaddrlen;
  char stmp[20];
  int i;

#ifdef RELAY_DEBUG
  struct in_addr addr;
  inet_aton("193.99.253.62", &addr);
/*  testconn(open_server(addr)); */
/*  open_server(addr); */
#endif

  while (1) {

    FD_ZERO(&rfds);
    FD_SET(guiservsock, &rfds);
    maxfd = guiservsock;
    if (guiconnected) {
      FD_SET(guiconnsock, &rfds);
      maxfd = MAX(maxfd, guiconnsock);
    }
    for (i=0; i<MAX_SERVERS; i++)
      if (sfstate[i]) {
        FD_SET(sfsocks[i], &rfds);
        maxfd = MAX(maxfd, sfsocks[i]);
      }

    selret = select(maxfd + 1, &rfds, NULL, NULL, NULL);

    get_time();

    if (selret <= 0) {
      if ((selret < 0) && (errno != EINTR)) {    /* error */
        fprintf(fdlog, "%s relay daemon terminated: fatal error in select call\n", lstime);
        fclose(fdlog);
        exit(0);
      }
      /* select terminated by signal */
      continue;
    }

    if (FD_ISSET(guiservsock, &rfds)) {

      tempaddrlen = sizeof(tempaddr);
      tempsock = accept(guiservsock, (struct sockaddr *) &tempaddr, &tempaddrlen);

      if (tempsock >= 0) {  /* accept successful */

        strcpy(stmp, inet_ntoa(tempaddr.sin_addr));

        for (i=0;
             (i < localaddrno) &&
             (tempaddr.sin_addr.s_addr != localaddr[i].s_addr);
             i++);

        if ((guiconnected) || (i == localaddrno)) {
          fprintf(fdlog, "%s GUI client connection refused: %s:%hu\n",
                  lstime, stmp, ntohs(tempaddr.sin_port));
          close(tempsock);
        }
        else {
          guiconnsock = tempsock;
          guicliaddr = tempaddr;
          guiconnected++;
          fprintf(fdlog, "%s GUI client connection established: %s:%hu\n",
                  lstime, stmp, ntohs(guicliaddr.sin_port));
        }
      }
      else
        fprintf(fdlog, "%s Failed accept on GUI client port\n", lstime);
    }

    for (i=0; i<MAX_SERVERS; i++)
      if ((sfstate[i]) && (FD_ISSET(sfsocks[i], &rfds)))
        handle_server(i);

    if ((guiconnected) && (FD_ISSET(guiconnsock, &rfds)))
      handle_client();

  }
}

/*
 * main()
 */

void main()
{
  int log;
#ifndef OS2
  struct passwd *pwd = NULL;
  int pid;

  /* don't run daemon as a super user */
  if ((getuid() == 0) || (geteuid() == 0))
    if (!(pwd = getpwnam(UNPRIV_NAME)) || (pwd->pw_uid <= 0)) {
      perror("Please create an unprivileged user named `"UNPRIV_NAME"'");
      exit(1);
    }
#else
  if (!RELAY_LOG) {
    fprintf(stderr, "Please set RELAY_LOG environment variable!\n");
    exit(1);
  }
#endif

  /* create log files and change owner if possible or if needed */
  if ((log = open(RELAY_LOG, O_CREAT | O_APPEND | O_RDWR, 0666)) < 0) {
    perror("Cannot open log file");
    exit(1);
  }
#ifndef OS2
  if ((getuid() == 0) || (geteuid() == 0))
    fchown(log, pwd->pw_uid, pwd->pw_gid);
#endif
  close(log);

#ifndef OS2
  if ((pid = fork()) == 0) { /* child */
    /* give up super user privileges */
    if ((getuid() == 0) || (geteuid() == 0)) {
      setgid(pwd->pw_gid);
      setuid(pwd->pw_uid);
    }
#endif

    /* become a daemon */
#ifndef OS2
    setsid();    /* disconnect from the terminal */
#endif
    chdir("/");  /* just in case someone wants to unmount the current dir */
    umask(077);

    get_time();

    initialize();

    main_loop(); /* this call should not return */

    exit(1);
#ifndef OS2
  }

  if (pid < 0) {
    perror("Cannot fork()");
    exit(1);
  }
#endif

  exit(0);

}

