
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.awt.*;
import java.util.*;
import java.io.*;
import Acme.Widgets.*;

/**
 * <b>List of IP addresses</b><br>
 * This class implements a list of IP addresses. The addresses are stored
 * in a vector of AddrObj objects. AddrObj is a private class declared in
 * this file. Each IP address consists of a four byte address and a four
 * byte netmask. This file also declares the private class AddrPanel which
 * is used to edit an address list.
 * @version 1.0 14 Dec 1996
 * @author Roland E. Schmid
 */

public class AddrList implements Persistent {

  /**
   * Initialize an empty address list 
   */
  public AddrList() {
    Addresses = new Vector(5,3);
  }

  /**
   * Add a new address to the address list. If the address already exists,
   * it is removed from the list before adding the new one. Two addresses
   * with equal address parts are equal regardless of their netmasks.
   * @param adr Address object to add
   */
  protected void addAddress(AddrObj adr) {
    Addresses.removeElement(adr);
    Addresses.addElement(adr);
  }

  /**
   * Add a new address to the address list. If the address already exists,
   * it is removed from the list before adding the new one. Two addresses
   * with equal address parts are equal regardless of their netmasks. The
   * given address and netmask are converted to an address object before
   * adding.
   * @param addr Byte array containing the IP address
   * @param mask Byte array containing the netmask
   */
  public void addAddress(byte[] addr, byte[] mask) {
    if (addr.length != 4 || mask.length != 4)
      return;
    AddrObj adr = new AddrObj(addr, mask);
    Addresses.removeElement(adr);
    Addresses.addElement(adr);
  }

  /**
   * Add a new address to the address list. If the address already exists,
   * it is removed from the list before adding the new one. Two addresses
   * with equal address parts are equal regardless of their netmasks. The
   * given string is converted to an address object before adding. The string
   * must be of the format <tt>"aaa.aaa.aaa.aaa[/mmm.mmm.mmm.mmm]"</tt>.
   * @param str string containing the IP address and optionally the netmask.
   */
  public void addAddress(String str) {
    AddrObj adr = new AddrObj(str);
    Addresses.removeElement(adr);
    Addresses.addElement(adr);
  }

  /** 
   * Delete an address from the address list. 
   * @param addr Byte array containing the address to delete.
   * @return true if successful, false if the parameter isn't a valid IP 
   * address or if the address is not in the list.
   */
  public boolean deleteAddress(byte[] addr) {
    if (addr.length != 4)
      return false;
    AddrObj adr = new AddrObj(addr, null);
    return Addresses.removeElement(adr);
  }

  /**
   * Create a graphical panel to edit the address list. The panel can
   * be included in frames or dialog boxes.
   * @return Panel containing the address list editor
   */
  public Panel editPanel() {
    return new AddrPanel(this);
  }

  /**
   * Get the first address of the address list. This method is called whenever
   * the software needs to know the address of a firewall it should use to
   * connect to.
   * @return IP address in byte array format or null if the list is empty
   */
  public byte[] getFirstAddress() {
    try {
      return ((AddrObj)Addresses.firstElement()).getAddr();
    }
    catch (NoSuchElementException e) {
      return null;
    }
  }

  /**
   * This method adds all addresses from the parameter address list,
   * that are not yet stored in the list. If both the list and the paramter
   * contain an address, the address with the shorter netmask is used.<br>
   * This method is used by the autoconfiguration algorithms.
   * @param al List containing the new addresses
   */
  public void union(AddrList al) {
    Enumeration en1;
    Enumeration en2 = al.getAddresses().elements();
    AddrObj ao1, ao2;
    byte ao1addr[], ao1mask[], ao2addr[], ao2mask[];
    outer_loop: while (en2.hasMoreElements()) { 
      ao2 = (AddrObj)en2.nextElement();
      ao2addr = ao2.getAddr();
      ao2mask = ao2.getMask();
      en1 = Addresses.elements();
      while (en1.hasMoreElements()) {
        ao1 = (AddrObj)en1.nextElement();
        ao1addr = ao1.getAddr();
        ao1mask = ao1.getMask();
        if (AddrObj.equalIP(AddrObj.And(ao1.AndMask(),ao2mask),
                            AddrObj.And(ao2.AndMask(),ao1mask))) {
          // addresses match
          if (AddrObj.maskLess(ao2mask, ao1mask)) { // new address more general, replace
            ao1.set(ao2addr, ao2mask);
          }
          continue outer_loop;
        }
      }
      // no match, add new address
      addAddress(ao2addr, ao2mask);
    }
  }

  /**
   * Get all addresses contained in the address list.
   * @return Vector containing the address objects
   */
  protected Vector getAddresses() {
    return Addresses;
  }

  /**
   * Check if the address list is empty
   * @return true, if the list is empty, false otherwise
   */
  public boolean empty() {
    return Addresses.isEmpty();
  }

  /**
   * Format all addresses contained in the list.<br>
   * The format of the strings is <tt>"aaa.aaa.aaa.aaa/mmm.mmm.mmm.mmm"</tt>
   * @return Array of strings containing the formatted addresses
   */ 
  public String[] printAddresses() {
    String str[] = new String[Addresses.size()];
    Enumeration en = Addresses.elements();
    AddrObj ao;
    int i=0;
    while (en.hasMoreElements()) {
      ao = (AddrObj)en.nextElement();
      str[i] = ao.toString();
      i++;
    }
    return str;
  }

  /**
   * Format all addresses and write them to the specified print stream.
   * The output format is conforming to the firewall configuration file syntax.
   * @param ps PrintStream to write the output to
   * @indent Number of spaces to prepend to each line
   * @port Port parameter to append to the addresses
   * @prend End value for a port range
   */
  public void printAddressesFormatted(PrintStream ps, int indent, int port, int prend) {
    String indentString = new String();
    for (int i=0; i < indent; i++)
      indentString += " ";

    String portString = new String();
    if (port != Macro.MACRO_FIRSTPORT || prend != Macro.MACRO_LASTPORT) {
      portString += " port "+port;
      if (port != prend)
        portString += " .. "+prend;
    }

    StringBuffer sb;
    Enumeration en = Addresses.elements();
    AddrObj ao;
    while (en.hasMoreElements()) {
      ao = (AddrObj)en.nextElement();
      sb = new StringBuffer(indentString);
      sb.append(Utils.printIP(ao.getAddr()));
      while (sb.length() < indent+15)
        sb.append(' ');
      sb.append(" mask ");
      sb.append(Utils.printIP(ao.getMask()));
      if (portString.length() > 0) {
        while (sb.length() < indent+36)
          sb.append(' ');
        sb.append(portString);
      }
      ps.print(sb.toString());
      if (en.hasMoreElements())
        ps.println(",");
      else
        ps.println();
    }
  }

  // Persistence
  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    ps.writePersistentVector("Addresses=", Addresses);
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    Addresses = ps.readPersistentVector("Addresses=");
  }

  private Vector Addresses;
}

// class to store an IP address together with its netmask
class AddrObj implements Persistent {

  // create an uninitialized address object, used when loading
  // data from an persistent input stream
  public AddrObj() {
  }

  // initialize object with given address and mask
  public AddrObj(byte[] addr, byte[] mask) {
    address = addr;
    netmask = mask;
    if (!equalZero(And(address,Not(netmask))))
      netmask = defaultMask(address);
  }

  // initialize object with the address given in str
  public AddrObj(String str) {
    int slashIndex = str.indexOf('/');
    String addrstr;
    if (slashIndex == -1)
      addrstr = str;
    else
      addrstr = str.substring(0, slashIndex);
    address = Communicator.resolve(addrstr);
    if (address == null)
      return;
    if (slashIndex != -1) {
      addrstr = str.substring(slashIndex + 1);
      netmask = Communicator.resolve(addrstr);
      if (!equalZero(And(address,Not(netmask)))) // host part of address is not zero
        netmask = null;                         // use default mask
      if (netmask == null) {
        netmask = new byte[4];
        slashIndex = -1; // use default mask
      }
    }
    else
      netmask = new byte[4];
    if (slashIndex == -1) 
      netmask = defaultMask(address);
  }

  // return bitwise AND of address and netmask
  public byte[] AndMask() {
    return And(address, netmask);
  }

  // return address
  public byte[] getAddr() {
    return address;
  }

  // return netmask
  public byte[] getMask() {
    return netmask;
  }

  // set address and netmask to given values
  public void set(byte addr[], byte mask[]) {
    address = addr;
    netmask = mask;
    if (!equalZero(And(address,Not(netmask))))
      netmask = defaultMask(address);
  }

  // compare to objects, ignoring the netmask (overrides Object.equals)
  public boolean equals(Object obj) { // ignore netmask
    byte[] cmp = ((AddrObj)obj).getAddr();
    return equalIP(address, cmp);
  }

  // convert object to string
  public String toString() {
    if (address == null)
      return "";
    StringBuffer stb = new StringBuffer(35);
    int j;
    for (j = 0; j < address.length; j++) {
      if (j != 0)
        stb.append(".");
      stb.append(Utils.unsign(address[j]));
    }
    if (netmask == null)
      return stb.toString();
    stb.append("/");
    for (j = 0; j < netmask.length; j++) {
      if (j != 0)
        stb.append(".");
      stb.append(Utils.unsign(netmask[j]));
    }
    return stb.toString();
  }

  // static methods

  // calculate default netmask
  public static byte[] defaultMask(byte a[]) {
    if (a.length != 4)
      return null;
    byte retval[] = new byte[4];
    long a0 = Utils.unsign(a[0]);
    retval[0] = -1; // 255
    if (a0 > 127)
      retval[1] = -1;
    if (a0 > 191)
      retval[2] = -1;
    if (a0 > 223)
      retval[3] = -1;
    if (!equalZero(And(a,Not(retval)))) // host address
      for (int i=0; i < 4; i++)
        retval[i] = -1;
    return retval;
  }

  // check if m1 is less than m2, i.e. m1 is more general than m2
  public static boolean maskLess(byte m1[], byte m2[]) {
    if (m1.length != 4 || m2.length != 4)
      return false;
    long l1, l2;
    for (int i=0; i < 4; i++) {
      l1 = Utils.unsign(m1[i]);
      l2 = Utils.unsign(m2[i]);
      if (l1 < l2)
        return true;
    }
    return false;
  }

  // bitwise AND of two IP addresses
  public static byte[] And(byte a1[], byte a2[]) {
    if (a1.length != 4 || a2.length != 4) 
      return null; 
    byte retval[] = new byte[4];
    for (int i = 0; i < 4; i++)
      retval[i] = (byte)(a1[i] & a2[i]);
    return retval;
  }

  // bitwise NOT of an IP address
  public static byte[] Not(byte a[]) {
    if (a.length != 4)
      return null;
    byte retval[] = new byte[4];
    for (int i=0; i < 4; i++)
      retval[i] = (byte)(~a[i]);
    return retval;
  }

  // test if two IP addresses are equal
  public static boolean equalIP(byte a1[], byte a2[]) {
    if (a1.length != a2.length)
      return false;
    for (int i=0; i < a1.length; i++)
      if (a1[i] != a2[i])
        return false;
    return true;
  }

  // test if IP address is zero
  public static boolean equalZero(byte a[]) {
    if (a.length != 4)
      return false;
    for (int i=0; i < 4; i++)
      if (a[i] != 0)
        return false;
    return true;
  }

  // Persistence
  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    int i;
    if (address == null)
      ps.writeInt("address=", 0);
    else {
      ps.writeInt("address=", address.length);
      for (i=0; i < address.length; i++)
        ps.writeInt(",", (int)address[i]);
    }
    if (netmask == null)
      ps.writeInt(",netmask=", 0);
    else {
      ps.writeInt(",netmask=", netmask.length);
      for (i=0; i < netmask.length; i++)
        ps.writeInt(",", (int)netmask[i]);
    }
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    int i, l;
    l = ps.readInt("address=");
    if (l == 0)
      address = null;
    else {
      address = new byte[l];
      for (i=0; i < l; i++)
        address[i] = (byte)ps.readInt(",");
    }
    l = ps.readInt(",netmask=");
    if (l == 0)
      netmask = null;
    else {
      netmask = new byte[l];
      for (i=0; i < l; i++)
        netmask[i] = (byte)ps.readInt(",");
    }
  }

  private byte[] address;
  private byte[] netmask;
}

// graphical panel to edit an address list
class AddrPanel extends Panel {

  public AddrPanel(AddrList al) {
    // implicit super() call

    AddrListObj = al;

    GridBagLayout gbl = new GridBagLayout();
    setLayout(gbl);
    
    addrlbox = new List(5, false);
    Enumeration en = al.getAddresses().elements();
    AddrObj adr;
    while (en.hasMoreElements()) {
      adr = (AddrObj)en.nextElement();
      addrlbox.addItem(adr.toString());
    }
    addbut = new Button("Add");
    removebut = new Button("Remove");
    addrtext = new TextField();
    addrlabel = new Label("Address:");
    masktext = new TextField();
    masklabel = new Label("Netmask:");

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    Utils.add_component(this, addrlbox, gbl, gbc, 0, 0, 4, 5, 100, 100);
    gbc.fill = GridBagConstraints.NONE;
    Utils.add_component(this, addbut, gbl, gbc, 2, 5, 1, 1, 0, 0);
    Utils.add_component(this, removebut, gbl, gbc, 1, 5, 1, 1, 0, 0);
    gbc.fill = GridBagConstraints.BOTH;
    Utils.add_component(this, addrlabel, gbl, gbc, 0, 6, 1, 1, 0, 0);
    Utils.add_component(this, addrtext, gbl, gbc, 1, 6, 3, 1, 100, 0);
    Utils.add_component(this, masklabel, gbl, gbc, 0, 7, 1, 1, 0, 0);
    Utils.add_component(this, masktext, gbl, gbc, 1, 7, 3, 1, 100, 0);
    addrtext.requestFocus();
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target == addbut) {
      String str;
      if (masktext.getText().trim().length() == 0)
        str = addrtext.getText().trim();
      else
        str = addrtext.getText().trim() + "/" + masktext.getText().trim();
      AddrObj ao = new AddrObj(str);
      if (ao.getAddr() == null) {
        ErrorBox eb = new ErrorBox(Utils.getParentFrame(this), 
          "Invalid IP address or host name:\n"+str);
        eb.show();
        return true;
      }
      if (AddrListObj.getAddresses().contains(ao)) {
        int sel = AddrListObj.getAddresses().indexOf(ao);
        addrlbox.delItem(sel);
      }
      AddrListObj.addAddress(ao);
      addrlbox.addItem(ao.toString());
      return true;
    }
    else if (evt.target == removebut) {
      int sel = addrlbox.getSelectedIndex();
      if (sel == -1) // nothing selected
        return true;
      AddrListObj.getAddresses().removeElementAt(sel);
      addrlbox.delItem(sel);
      return true;
    }
    else return super.action(evt, arg);
  }

  private AddrList AddrListObj;
  private List addrlbox;
  private Button addbut;
  private Button removebut;
  private TextField addrtext;
  private TextField masktext;
  private Label addrlabel;
  private Label masklabel;
}
