
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class implements the communication with the relay. All methods
 * in this class are static and synchronized.
 * @version 1.0 10 Dec 1996
 * @author Roland E. Schmid
 */

public class Communicator {

  static { // This is a static initializer
    // connect to relay on local host
    connectRelay();
  }

  // public methods

  /**
   * Open server connection
   * @param servAddr server address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean openServer(byte[] servAddr) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    openConns.removeElement(servAddr); // removes element if it exists
    try {
      writeHeader(CLI_OPEN_SERV, 4);
      out.write(servAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
        case CLI_ERROR_ISCONN:
          openConns.addElement(servAddr);
          return true;
        case CLI_ERROR_OPENCONN:
          commError = COMM_OPENCONN;
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Close server connection
   * @param servAddr server address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean closeServer(byte[] servAddr) {
    if (!openConns.contains(servAddr)) 
      return true;
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    try {
      writeHeader(CLI_CLOSE_SERV, 4);
      out.write(servAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
        case CLI_ERROR_NOTCONN:
          openConns.removeElement(servAddr);
          return true;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      }
    }
    catch (Exception e) {
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Check generated config file for errors and rename it to firewall.conf
   * @param servAddr Server address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean checkConfig(byte[] servAddr) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      writeHeader(CLI_CHECKCONFIG, 4);
      out.write(servAddr, 0, 4);
      readHeader();
      if (readCommand != CLI_OK) {
        if (readCommand == CLI_ERROR_NOTCONN)
          openServer(servAddr);
        commError = COMM_ERROR;
        in.skipBytes(readLength);
        return false;
      }
      readHeader(); // read result of check operation
      return (readCommand == CLI_OK);
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Reconfigure the firewall
   * @param servAddr Server Address
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean reconfig(byte[] servAddr) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
        writeHeader(CLI_RECONFIG, 4);
        out.write(servAddr, 0, 4);
        readHeader();
        if (readCommand != CLI_OK) {
          if (readCommand == CLI_ERROR_NOTCONN) {
            if (openServer(servAddr))
              continue;
            else
              return false;
          }
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
        }
        readHeader(); // read result of check operation
        if (readCommand == CLI_OK) {
          try {
            Thread.sleep(10000);
          }
          catch (InterruptedException ignore) {}
          return openServer(servAddr);
        }
        else
          return false;
      } // for
      return false;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Fetch file from server and store as temporary file
   * @param servAddr server address
   * @param fname filename
   * @return filename of temporary file, commError is set when returning null
   */
  public static synchronized String getTempFile(byte[] servAddr, String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return null;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return null;
    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
  	String filename = fname.trim() + "\u0000";
    	int len = 4 + filename.length();
      	writeHeader(CLI_GET_FILE_TMP, len);
  	out.write(servAddr, 0, 4);
    	out.writeBytes(filename);
      	readHeader();
	switch (readCommand) {
          case CLI_FILE_TMP_NAME:
            byte tname[] = new byte[readLength];
            in.readFully(tname);
            readHeader();
            if (readCommand == CLI_OK)
              return new String(tname, 0, 0, tname.length-1);
            commError = COMM_ERROR;
            in.skipBytes(readLength);
            return null;
          case CLI_ERROR_NOTCONN:
            if (!openServer(servAddr)) {
              return null;
            }
            break;
          case CLI_ERROR_FILE:
            commError = COMM_FILEERR;
            return null;
          default:
            commError = COMM_ERROR;
            in.skipBytes(readLength);
            return null;
        } // switch
      } // for
      dispose();
      commError = COMM_IO;
      return null;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return null;
    }
  }

  /**
   * Request file from server. This method is called from RelayInputStream
   * @param servAddr server address
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see RelayInputStream
   */
  public static synchronized boolean requestFile(byte[] servAddr, String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
        String filename = fname.trim() + "\u0000";
        int len = 4 + filename.length();
        writeHeader(CLI_GET_FILE, len);
        out.write(servAddr, 0, 4);
        out.writeBytes(filename);
        readHeader();
        switch (readCommand) {
          case CLI_OK:
            return true;
          case CLI_ERROR_NOTCONN:
            if (!openServer(servAddr)) {
              return false;
            }
            break;
          default:
            commError = COMM_ERROR;
            in.skipBytes(readLength);
            return false;
        } // switch
      } // for
      dispose();
      commError = COMM_IO;
      return false;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Request local file from relay. This method is called from LocalInputStream
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see LocalInputStream
   */
  public static synchronized boolean requestLocalFile(String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    try {
      String filename = fname.trim() + "\u0000";
      int len = filename.length();
      writeHeader(CLI_GET_LOCAL_FILE, len);
      out.writeBytes(filename);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          return true;
        case CLI_ERROR_FILE:
          commError = COMM_FILEERR;
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Write file to server. This method is called from RelayOutputStream
   * @param servAddr server address
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see RelayOutputStream
   */
  public static synchronized boolean writeFile(byte[] servAddr, String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      for (int i = 0; i < 3; i++) { // max. 3 retries when ERROR_NOTCONN
	String filename = fname.trim() + "\u0000";
	int len = 4 + filename.length();
	writeHeader(CLI_WRITE_FILE, len);
	out.write(servAddr, 0, 4);
	out.writeBytes(filename);
	readHeader();
	switch (readCommand) {
          case CLI_OK:
            return true;
          case CLI_ERROR_NOTCONN:
            if (!openServer(servAddr)) {
              return false;
            }
            break;
          default:
            commError = COMM_ERROR;
            in.skipBytes(readLength);
            return false;
        } // switch
      } // for
      dispose();
      commError = COMM_IO;
      return false;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Write local file to relay. This method is called from LocalOutputStream
   * @param fname filename
   * @return true if successful, commError is set when returning false
   * @see LocalOutputStream
   */
  public static synchronized boolean writeLocalFile(String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    try {
      String filename = fname.trim() + "\u0000";
      int len = filename.length();
      writeHeader(CLI_WRITE_LOCAL_FILE, len);
      out.writeBytes(filename);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          return true;
        case CLI_ERROR_FILE:
          commError = COMM_FILEERR;
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Get temporary local filename from relay. This method is used to
   * allocate a temporary file name before creating a local file using
   * the LocalOutputStream.
   * @return filename, commError is set when returning null
   */
  public static synchronized String getTempName() {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return null;
      }
    try {
      writeHeader(CLI_GET_TEMPNAME, 0);
      readHeader();
      switch (readCommand) {
        case CLI_FILE_TMP_NAME:
          byte tname[] = new byte[readLength];
          in.readFully(tname);
          return new String(tname, 0, 0, tname.length-1);
        case CLI_ERROR_FILE:
          commError = COMM_FILEERR;
          return null;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return null;
      } // switch
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return null;
    }
  }

  /**
   * Delete local file
   * @param fname filename
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean deleteFile(String fname) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    try {
      String filename = fname.trim() + "\u0000";
      writeHeader(CLI_DELETE_FILE, filename.length());
      out.writeBytes(filename);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          return true;
        case CLI_ERROR_FILE:
          commError = COMM_FILEERR;
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Read active TCP connections from firewall
   * @param host Firewall to read configuration from
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean readTcpConns(Host host) {
    int tmpint;
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!host.isFirewall) {
      commError = COMM_NOFIREWALL;
      return false;
    }
    byte[] hostAddr = host.HostAddresses.getFirstAddress();
    if (hostAddr == null) {
      commError = COMM_NOADDRESS;
      return false;
    }
    if (!openConns.contains(hostAddr))
      if (!openServer(hostAddr))
        return false;
    try {
      writeHeader(CLI_READ_TCP, 4);
      out.write(hostAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          break;
        case CLI_ERROR_NOTCONN:
          openServer(hostAddr);
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
      TcpConnection tc;
      host.tcpconns = new Vector(10,10);
      readHeader();
      if (readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      tmpint = in.readInt();
      host.actDate = new Date((long)tmpint * 1000);
      readHeader();
      while (readLength == 48) {
        tc = new TcpConnection();
        host.tcpconns.addElement(tc);
        tc.id = in.readInt();
        tc.key = in.readInt();
        tmpint = in.readInt();
        tc.created = new Date((long)tmpint * 1000);
        tc.created_jiffies = in.readInt();
        tmpint = in.readInt();
        tc.lastuse = new Date((long)tmpint * 1000);
        tmpint = in.readInt();
        tc.timeout = new Date((long)tmpint * 1000);
        tc.fromaddr = new byte[4];
        in.readFully(tc.fromaddr, 0, 4);
        tc.fromport = in.readInt();
        tc.toaddr = new byte[4];
        in.readFully(tc.toaddr, 0, 4);
        tc.toport = in.readInt();
        tmpint = in.readInt();
        tc.allpsource = tmpint != 0;
        tc.state = in.readInt();
        readHeader();
      }
      if (readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }
      return true;
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Kill TCP connection on firewall
   * @param servAddr Server address (firewall)
   * @param id TCP connection ID
   * @param key TCP connection key
   * @param created Creation time of TCP connection
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean killTcp(byte[] servAddr, 
                                       int id, int key, int created) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      writeHeader(CLI_KILL_TCP, 16);
      out.write(servAddr, 0, 4);
      out.writeInt(id);
      out.writeInt(key);
      out.writeInt(created);
      readHeader();
      if (readCommand != CLI_OK) {
        if (readCommand == CLI_ERROR_NOTCONN)
          openServer(servAddr);
        commError = COMM_ERROR;
        in.skipBytes(readLength);
        return false;
      }
      return true;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Reset usage counter on firewall
   * @param servAddr Server address (firewall)
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean resetUsage(byte[] servAddr) {
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!openConns.contains(servAddr))
      if (!openServer(servAddr))
        return false;
    try {
      writeHeader(CLI_RESET_USAGE, 4);
      out.write(servAddr, 0, 4);
      readHeader();
      if (readCommand != CLI_OK) {
        if (readCommand == CLI_ERROR_NOTCONN)
          openServer(servAddr);
        commError = COMM_ERROR;
        in.skipBytes(readLength);
        return false;
      }
      return true;
    } // try
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    }
  }

  /**
   * Read active firewall configuration
   * @param host Firewall to read configuration from
   * @return true if successful, commError is set when returning false
   */
  public static synchronized boolean readFwConfig(Host host) {
    int tmpint;
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return false;
      }
    if (!host.isFirewall) {
      commError = COMM_NOFIREWALL;
      return false;
    }
    byte[] hostAddr = host.HostAddresses.getFirstAddress();
    if (hostAddr == null) {
      commError = COMM_NOADDRESS;
      return false;
    }
    if (!openConns.contains(hostAddr))
      if (!openServer(hostAddr))
        return false;
    try {
      writeHeader(CLI_READ_ADDR, 4);
      out.write(hostAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          break;
        case CLI_ERROR_NOTCONN:
          openServer(hostAddr);
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
      readHeader();
      if (readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      int numEntries = in.readInt();
      host.addr = new byte[numEntries][4];
      host.mask = new byte[numEntries][4];
      host.port = new short[numEntries];
      host.prend = new short[numEntries];
      int pos = 0;
      readHeader();
      while (readLength == 12) {
        in.readFully(host.addr[pos], 0, 4);
        in.readFully(host.mask[pos], 0, 4);
        host.port[pos] = in.readShort();
        host.prend[pos] = in.readShort();
        readHeader();
        pos++;
      }
      if (readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      writeHeader(CLI_READ_RULES, 4);
      out.write(hostAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          break;
        case CLI_ERROR_NOTCONN:
          openServer(hostAddr);
          return false;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
      host.rules = new Vector(20,20);
      Rule r;
      readHeader();
      while (readLength == 64) {
        r = new Rule();
        r.fw_line = in.readInt();
        r.level_num = in.readInt();
        r.timeout = null;
        tmpint = in.readInt();
        if (tmpint < 0)
          r.timeout_type = tmpint;
        else {
          r.timeout_type = Rule.SF_TIMEOUT_VALID;
          r.timeout = new Date((long)tmpint * 1000);
        }
        r.fw_rc = in.readInt();
        r.fw_src_idx = in.readInt();
        r.fw_src_cnt = in.readInt();
        r.fw_dst_idx = in.readInt();
        r.fw_dst_cnt = in.readInt();
        r.fw_flags = in.readInt();
        r.ttl = in.readInt();
        r.protocol = in.readInt();
        r.fw_rip_idx = in.readInt();
        r.fw_rip_cnt = in.readInt();
        r.usage = in.readInt();
        r.bytes = (long)in.readInt();
        tmpint = in.readInt();
        r.bytes += (long)tmpint * Rule.GIGA;
        host.rules.addElement(r);
        readHeader();
      }
      if (readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      writeHeader(CLI_READ_VARS, 4);
      out.write(hostAddr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_OK:
          break;
        case CLI_ERROR_NOTCONN:
          openServer(hostAddr);
          break;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return false;
      } // switch
      // read number of variables
      readHeader();
      if (readLength != 4) {
        commError = COMM_ERROR;
        return false;
      }
      int numVars = in.readInt();
      int varcnt = 0;
      host.varnames = new String[numVars];
      host.varvalues = new int[numVars][];
      host.varaddr = new byte[numVars][][];
      host.vartimeout = new Date[numVars][];
      readHeader();
      byte buf[] = new byte[256];
      while (readLength != 0) {
        // read variable name
        in.readFully(buf, 0, readLength);
        host.varnames[varcnt] = new String(buf, 0, 0, readLength);
        
        // read number of entries
        readHeader();
        if (readLength != 4) {
          commError = COMM_ERROR;
          return false;
        }
        numEntries = in.readInt();
        host.varvalues[varcnt] = new int[numEntries];
        host.varaddr[varcnt] = new byte[numEntries][4];
        host.vartimeout[varcnt] = new Date[numEntries];

        for (int i=0; i < numEntries; i++) {
          readHeader();
          if (readLength != 12) {
            commError = COMM_ERROR;
            return false;
          }
          host.varvalues[varcnt][i] = in.readInt();
          in.readFully(host.varaddr[varcnt][i], 0, 4);
          tmpint = in.readInt();
          host.vartimeout[varcnt][i] = new Date((long)tmpint * 1000);
        }

        readHeader();
        varcnt++;
      }
      if (readLength != 0) {
        commError = COMM_ERROR;
        return false;
      }

      return true; 
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return false;
    } 
  }

  /**
   * Resolve host name
   * @param hname hostname or string containing IP address
   * @return IP address if successful, commError is set when returning null
   */
  public static synchronized byte[] resolve(String hname) {
    try {
      return Utils.parseIP(hname);
    }
    catch (IllegalArgumentException e) {}
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return null; 
      }
    try {
      String filename = hname.trim() + "\u0000";
      writeHeader(CLI_RESOLVE, filename.length());
      out.writeBytes(filename);
      readHeader();
      switch (readCommand) {
        case CLI_RESOLVE_REPLY:
          if (readLength == 0) {
            commError = COMM_RESOLVE;
            return null;
          }
          if (readLength == 4) {
            byte tempaddr[] = new byte[4];
            in.read(tempaddr, 0, 4);
            return tempaddr;
          }
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return null;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return null;
    }
  }

  /**
   * Resolve IP address
   * @param addr IP address
   * @return hostname or IP address if not resolvable
   */
  public static synchronized String resolveIP(byte[] addr) {
    String key = Utils.printIP(addr);
    String hostname = (String)IPCache.get(key);
    if (hostname != null) {
      return hostname;
    }
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return key; 
      }
    try {
      writeHeader(CLI_RESOLVE_IP, 4);
      out.write(addr, 0, 4);
      readHeader();
      switch (readCommand) {
        case CLI_RESOLVE_REPLY:
          if (readLength == 0) {
            IPCache.put(key, key);
            return key;
          }
          byte hname[] = new byte[readLength];
          in.readFully(hname);
          hostname = new String(hname, 0, 0, hname.length-1);
          IPCache.put(key, hostname);
          return hostname;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return key;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return key;
    }
  }

  /**
   * Resolve service name
   * @param sname service name
   * @param proto protocol
   * @return service number if successful, -1 otherwise
   */
  public static synchronized int resolveService(String sname, String proto) {
    try {
      int i = Integer.parseInt(sname);
      if (i>=Macro.MACRO_FIRSTPORT && i<=Macro.MACRO_LASTPORT)
        return i;
    }
    catch (NumberFormatException ignore) {}
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return -1; 
      }
    try {
      String servname = sname.trim() + "\u0000";
      String protname = proto.trim() + "\u0000";
      writeHeader(CLI_RESOLVE_SERVICE, servname.length()+protname.length());
      out.writeBytes(servname);
      out.writeBytes(protname);
      readHeader();
      switch (readCommand) {
        case CLI_RESOLVE_REPLY:
          if (readLength == 0) {
            commError = COMM_RESOLVE;
            return -1;
          }
          if (readLength == 4) {
            int portno = in.readInt();
            return portno;
          }
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return -1;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return -1;
    }
  }

  /**
   * Resolve service number
   * @param port service number
   * @param proto protocol
   * @param privOnly resolve only if port < 1024
   * @return service name or number if not resolvable
   */
  public static synchronized String resolvePort(int port, 
    String proto, boolean privOnly) {
    if (port == 7227)
      return "firewall";
    if (port == relayPort)
      return "sfrelay";
    String portString = Integer.toString(port);
    if (privOnly && port >= 1024)
      return portString;
    Hashtable portCache = proto.toUpperCase().equals("UDP") ? portCacheUDP : portCacheTCP;
    String portName = (String)portCache.get(portString);
    if (portName != null) {
      return portName;
    }
    if (!connected)
      if (!connectRelay()) {
        commError = COMM_NOCONN;
        return portString; 
      }
    try {
      String protname = proto.trim() + "\u0000";
      writeHeader(CLI_RESOLVE_PORT, 4+protname.length());
      out.writeInt(port);
      out.writeBytes(protname);
      readHeader();
      switch (readCommand) {
        case CLI_RESOLVE_REPLY:
          if (readLength == 0) {
            portCache.put(portString, portString);
            return portString;
          }
          byte hname[] = new byte[readLength];
          in.readFully(hname);
          portName = new String(hname, 0, 0, hname.length-1);
          portCache.put(portString, portName);
          return portName;
        default:
          commError = COMM_ERROR;
          in.skipBytes(readLength);
          return portString;
      }
    }
    catch (Exception e) {
      dispose();
      commError = COMM_IO;
      return portString;
    }
  }

  /**
   * Close connection to sfrelay and all server connections
   */
  public static synchronized void dispose() {
    if (connected)
      try {
        Enumeration servers = openConns.elements();
        while (servers.hasMoreElements())
          closeServer((byte[])servers.nextElement());
        iosock.close();
      }
      catch (Exception e) {}
    connected = false;
  }

  /**
   * Establish connection to sfrelay
   * @return true if successful
   */
  public static synchronized boolean connectRelay() {
    if (connected)
      return true;
    try {
      InetAddress localaddr = InetAddress.getLocalHost();
      iosock = new Socket(localaddr, relayPort);
      in = new DataInputStream(iosock.getInputStream());
      out = new DataOutputStream(iosock.getOutputStream());
      connected = true;
      return true;
    }
    catch (UnknownHostException e) {
      System.out.println("Error: Cannot get local IP address!");
    }
    catch (IOException e) {
      System.out.println("Error: Unable to connect to sfrelay!");
    }
    catch (Exception e) {
      System.out.println("Error: Unexpected error connecting to sfrelay!");
    }
    return false;
  }

  /**
   * Write header to output stream. This methods writes the packet header
   * of the control-panel-to-relay protocol to the output stream.
   * @param command Command for the relay
   * @param length Number of bytes following the header
   */
  protected static synchronized void writeHeader(int command, int length) throws IOException {
    out.writeInt(magic);
    out.writeInt(command);
    out.writeInt(length);
  }

  /**
   * Read header from output stream. The read command is stored in readCommand and
   * the number of bytes following the header is stored in readLength.
   */
  protected static synchronized void readHeader() throws IOException, EOFException {
    commError = COMM_IO; // If an exception is thrown here, it is probably an I/O error
    int mg = in.readInt();
    if (mg != magic)
      throw new IOException("Invalid magic");
    readCommand = in.readInt();
    readLength = in.readInt();
    commError = COMM_OK;
  }

  // public data

  /**
   * Value of commError: successful operation
   */
  public final static int COMM_OK          =    0;
  /**
   * Value of commError: unspecified error
   */
  public final static int COMM_ERROR       =   -1;
  /**
   * Value of commError: cannot connect to relay
   */
  public final static int COMM_NOCONN      =   -2;
  /**
   * Value of commError: general I/O error (e.g. connection aborted)
   */
  public final static int COMM_IO          =   -3; 
  /**
   * Value of commError: relay cannot connect to server
   */
  public final static int COMM_OPENCONN    =   -4; 
  /**
   * Value of commError: error handling a local file
   */
  public final static int COMM_FILEERR     =   -5; 
  /**
   * Value of commError: error resolving hostname, IP address, service name
   * or port number
   */
  public final static int COMM_RESOLVE     =   -6; 
  /**
   * Value of commError: The parameter passed to the method does not refer
   * to a firewall.
   */
  public final static int COMM_NOFIREWALL  =   -7;
  /**
   * Value of commError: The host object passed to the method does not
   * contain an IP address.
   */
  public final static int COMM_NOADDRESS   =   -8;

  /**
   * The variable commError holds the error code of the last operation
   */
  public static int commError = COMM_OK; // errorcode for last command

  /**
   * TCP port on which the relay is accepting connections
   */
  protected final static int relayPort = 7228;

  // magic used in relay communication
  private final static int magic = 0xbadcafe; 

  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_OPEN_SERV      = 0x0100;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_CLOSE_SERV     = 0x0200;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_GET_FILE_TMP   = 0x0300; /* filename must include full path */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_FILE_TMP_NAME  = 0x0400;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_GET_FILE       = 0x0500; /* filename must include full path */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_FILE_DATA      = 0x0600;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_DELETE_FILE    = 0x0700;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESOLVE        = 0x0800;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESOLVE_REPLY  = 0x0900;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_WRITE_FILE     = 0x0A00; /* expects filename without path ! */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_WRITE_STATUS   = 0x0B00;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_READ_ADDR	= 0x0C00;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_READ_RULES	= 0x0D00;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_READ_VARS	= 0x0E00;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_WRITE_LOCAL_FILE = 0x0F00;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RECONFIG       = 0x1000;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_CHECKCONFIG    = 0x1100;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_READ_TCP       = 0x1200;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_KILL_TCP       = 0x1300;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESET_USAGE    = 0x1400;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_GET_LOCAL_FILE = 0x1500;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_GET_TEMPNAME   = 0x1600;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESOLVE_IP     = 0x1700;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESOLVE_SERVICE= 0x1800;
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_RESOLVE_PORT   = 0x1900;

  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_OK             = 0xA100; /* no error */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_ERROR          = 0xA200; /* unspecified error */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_ERROR_ISCONN   = 0xA300; /* already connected to server */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_ERROR_OPENCONN = 0xA400; /* cannot open connection to server */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_ERROR_NOTCONN  = 0xA500; /* not connected to server */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_ERROR_FILE     = 0xA600; /* error handling local file */
  /** 
   * Command used in communication with the relay. 
   */
  protected final static int CLI_WAIT		= 0xA700; /* write status not yet available */

  /**
   * Input stream to read from relay
   */
  protected static DataInputStream in;
  /**
   * Output stream to write to relay;
   */
  protected static DataOutputStream out;
  /**
   * The last command read from the relay is saved in readCommand.
   */
  protected static int readCommand;		// last command read from in
  /**
   * The last packet data length read from the relay is saved in readLength.
   */
  protected static int readLength;		// last data length read from in

  // private data
  private static boolean connected = false;	           // connected to sfrelay
  private static Socket iosock;			           // socket
  private static Vector openConns = new Vector(5,5);       // open server connections
  private static Hashtable portCacheUDP = new Hashtable(); // Cache for UDP port translation
  private static Hashtable portCacheTCP = new Hashtable(); // Cache for TCP port translation
  private static Hashtable IPCache = new Hashtable();      // Cache for IP address translation

} // Communicator

