
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.awt.*;

/**
 * Interface for drag-and-drop objects used by class DragDrop
 * @see DragDrop
 * @version 1.0 29 Nov 1996
 * @author Roland E. Schmid
 */

public interface DragDropObj {

  /**
   * Load image if not yet loaded<br>
   * This function is called by DragDrop immediately after object creation.
   * @param target component where image is to be drawn
   * @param mt MediaTracker: DragDrop waits for all images in mt to be loaded
   */
  public void initImage(Component target, MediaTracker mt);

  /**
   * Determine if the mouse cursor is located above the object.
   * @param x virtual x coordinate
   * @param y virtual y coordinate
   * @return true if the object covers (x,y)
   */
  public boolean hasFocus(int x, int y);

  /**
   * Set new virtual object coordinates.
   * This should be called after erase() and before draw()
   * @param x virtual x coordinate
   * @param y virtual y coordinate
   * @see #draw
   * @see #erase
   */
  public void setCoordinates(int x, int y);

  /**
   * Get x coordinate.
   * @return x coordinate
   */
  public int getX();

  /**
   * Get y coordinate.
   * @return y coordinate
   */
  public int getY();

  /**
   * Get coordinates for connecting an edge towards destination.
   * @param destination coordinates of destination vertex
   * @return coordinates on the border of the object
   */
  public Point connectionCoord(Point destination);

  /**
   * Draw object at position set with setCoordinates().
   * @param g Graphics object for drawing routines
   * @param x_offset virtual x offset of canvas origin (due to scrolling)
   * @param y_offset virtual y offset of canvas origin (due to scrolling)
   * @see #setCoordinates
   */
  public void draw(Graphics g, int x_offset, int y_offset);

  /**
   * Erase object at position set with setCoordinates().
   * @param g Graphics object for drawing routines
   * @param x_offset virtual x offset of canvas origin (due to scrolling)
   * @param y_offset virtual y offset of canvas origin (due to scrolling)
   * @see #setCoordinates
   */
  public void erase(Graphics g, int x_offset, int y_offset);

  /**
   * Turn object highlight on or off.
   * @param g Graphics object for drawing routines
   * @param x_offset virtual x offset of canvas origin (due to scrolling)
   * @param y_offset virtual y offset of canvas origin (due to scrolling)
   * @param highlight true: turn highlight on, false: turn highlight off
   */
  public void highlight(Graphics g, int x_offset, int y_offset, boolean highlight);

  /**
   * Check if object is highlighted.
   * @return true if object is highlighted, false otherwise
   */
  public boolean isHighlight();

  /**
   * Execute object specific routines, e.g. property setting dialog. 
   * execute() is called when the user double-clicks on the object and
   * edit mode is enabled.
   * @param parent Parent frame for execute() to be able to use dialog
   * boxes or to create child frames.
   */
  public void execute(Frame parent);

  /**
   * Called before deleting an object.
   * @return true if object deletion may proceed
   */
  public boolean confirmDelete();

  /**
   * Called when editing is disabled and user clicks on an object.
   * @param g Graphics object for drawing routines
   * @param x_offset virtual x offset of canvas origin (due to scrolling)
   * @param y_offset virtual y offset of canvas origin (due to scrolling)
   */
  public void userAction(Frame parent, Graphics g, int x_offset, int y_offset);

  /**
   * Returns a String describing the object. This string is displayed in
   * the status line when the object is selected.
   * @return objectID string
   */
  public String getObjectID();

}

