
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.util.*;

/**
 * This class implements a graph datastructure.
 * @version 1.0 29 Nov 1996
 * @author Roland E. Schmid
 */

public class Graph implements Persistent {

  /**
   * @return true iff the graph does not contain any vertices
   */
  public boolean isEmpty() {
    return Vertices.isEmpty();
  }

  /**
   * @return number of vertices in the graph
   */
  public int NumberOfVertices() {
    return Vertices.size();
  }

  /**
   * Search a vertex in the graph
   * @param V the object to search 
   * @return true if the object is a vertex in the graph
   */
  public boolean searchVertex(Object V) {
    return Vertices.contains(V);
  }
    
  /**
   * Search an edge in the graph
   * @param V1 first vertex
   * @param V2 second vertex
   * @return true if V1 and V2 are adjacent
   */
  public boolean searchEdge(Object V1, Object V2) {
    int V1index = Vertices.lastIndexOf(V1);
    if (V1index == -1) 
      return false;
    return ((Vector)(AdjLists.elementAt(V1index))).contains(V2);
  }

  /**
   * Insert a vertex in graph
   * @param V object to insert
   */
  public void insertVertex(Object V) {
    if (!Vertices.contains(V)) {
      Vertices.addElement(V);
      Vector al = new Vector(3,3);
      AdjLists.addElement(al);
    }
  }

  /**
   * Insert an edge in graph
   * @param V1 first vertex
   * @param V2 second vertex
   */
  public void insertEdge(Object V1, Object V2) {
    if ((!Vertices.contains(V1)) || (!Vertices.contains(V2)))
      return;
    Vector al = (Vector)(AdjLists.elementAt(Vertices.lastIndexOf(V1)));
    if (!al.contains(V2))
      al.addElement(V2);
    al = (Vector)(AdjLists.elementAt(Vertices.lastIndexOf(V2)));
    if (!al.contains(V1))
      al.addElement(V1);
  }

  /**
   * Delete a vertex
   * @param V vertex to delete
   */
  public void deleteVertex(Object V) {
    int index = Vertices.lastIndexOf(V);
    if (index == -1)
      return;
    Vertices.removeElementAt(index);
    AdjLists.removeElementAt(index);
    Enumeration alenum = AdjLists.elements();
    while (alenum.hasMoreElements()) {
      Vector al = (Vector)alenum.nextElement();
      al.removeElement(V);
    }
  }

  /**
   * Delete an edge
   * @param V1 first vertex
   * @param V2 second vertex
   */
  public void deleteEdge(Object V1, Object V2) {
    if ((!Vertices.contains(V1)) || (!Vertices.contains(V2)))
      return;
    Vector al = (Vector)(AdjLists.elementAt(Vertices.lastIndexOf(V1)));
    al.removeElement(V2);
    al = (Vector)(AdjLists.elementAt(Vertices.lastIndexOf(V2)));
    al.removeElement(V1);
  }

  /**
   * @param V vertex
   * @return enumeration of all neighbors of vertex V
   */
  public Enumeration findNeighbors(Object V) {
    return new GraphEnumeration(Vertices, AdjLists, V);
  }

  /**
   * @return an enumeration of all vertices in graph
   */
  public Enumeration getAllVertices() {
    return new GraphEnumeration(Vertices, AdjLists);
  }

  /**
   * @return an enumeration of all edges in graph
   * @see Edge
   */
  public Enumeration getAllEdges() {
    return new EdgeEnumeration(Vertices, AdjLists);
  }

  /**
   * override Object.toString()
   * @return string describing the graph
   */
  public String toString() {
    String str = new String("number of vertices: "+Vertices.size());
    Vector al;
    for (int i=0; i < Vertices.size(); i++) {
      str += "\n  "+i+" "+Vertices.elementAt(i)+"\n     Neighbors:";
      al = (Vector)AdjLists.elementAt(i);
      for (int j=0; j < al.size(); j++)
        str += " "+Vertices.lastIndexOf(al.elementAt(j));
    }  
    return str;
  }

  // Persistence
  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    ps.writePersistentVector("vertices=", Vertices);
    for (int i = 0; i < Vertices.size(); i++)
      ps.writePersistentVector(",adjlist=", (Vector)AdjLists.elementAt(i));
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    Vertices = ps.readPersistentVector("vertices=");
    for (int i = 0; i < Vertices.size(); i++) 
      AdjLists.addElement(ps.readPersistentVector(",adjlist="));
  }

  // private data
  private Vector Vertices = new Vector(10, 5);
  private Vector AdjLists = new Vector(10, 5);
}

// Enumeration of all vertices
class GraphEnumeration implements Enumeration {

  // constructors
  public GraphEnumeration(Vector Vertices, Vector AdjLists) {
    selectedVertices = Vertices.elements();
  }

  public GraphEnumeration(Vector Vertices, Vector AdjLists, Object Neighbor) {
    int index = Vertices.lastIndexOf(Neighbor);
    if (index == -1)
      NeighborAvailable = false;
    else 
      selectedVertices = ((Vector)(AdjLists.elementAt(index))).elements();
  }

  // public instance methods
  public boolean hasMoreElements() {
    if (NeighborAvailable)
      return selectedVertices.hasMoreElements();
    else
      return false;
  }

  public Object nextElement() {
    if (NeighborAvailable)
      return selectedVertices.nextElement();
    else
      throw new NoSuchElementException();
  }

  // private data
  private Enumeration selectedVertices;
  private boolean NeighborAvailable = true;
}

// Enumeration of all edges
class EdgeEnumeration implements Enumeration {

  // constructor
  public EdgeEnumeration(Vector Vertices, Vector AdjLists) {
    Edge e;
    for (int i = 0; i < Vertices.size(); i++) {
      Vector al = (Vector)AdjLists.elementAt(i);
      for (int j = 0; j < al.size(); j++) {
        Object o = al.elementAt(j);
        if (Vertices.lastIndexOf(o) > i) { // insert each edge only once
          e = new Edge(Vertices.elementAt(i), o);
          Edges.addElement(e);
        }
      }
    }
    EdgeEnum = Edges.elements();
  }

  // public instance methods
  public boolean hasMoreElements() {
    return EdgeEnum.hasMoreElements();
  }

  public Object nextElement() {
    return EdgeEnum.nextElement();
  }

  // private data
  private Vector Edges = new Vector(20,10);
  private Enumeration EdgeEnum;
}
