
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.awt.*;
import java.net.*;
import java.util.*;
import java.io.*;
import Acme.Widgets.*;

/**
 * This class implements host objects. It describes the properties
 * of general hosts, firewalls and servers. 
 * @version 1.0 03 Dec 1996
 * @author Roland E. Schmid
 */

public class Host extends IconObject {

  /**
   * Initialize host and host icon
   */
  public Host() {
    // implicit call to super()
    if (bSource == null)
      bSource = new cpu1ImageSource();
    icon = Toolkit.getDefaultToolkit().createImage(bSource);

    hostTemplates = new boolean[Templates.numberServer + Templates.numberFirewall];
    for (int i=0; i < hostTemplates.length; i++)
      hostTemplates[i] = Templates.defaultActivate[i+Templates.startServer];
  }

  /**
   * Load the actual configuration data from the firewall
   * @return true if successful, false otherwise
   */
  public boolean loadDynamicConfig() {
    isLoaded = false;
    if (!Communicator.readFwConfig(this)) {
      if (Communicator.commError == Communicator.COMM_OK) {
        // retry once, server connection may be reestablished now
        if (!Communicator.readFwConfig(this))
          return false;
      }
      else
        return false;
    }
    isLoaded = true;
    return true;
  }

  /**
   * Load the active TCP connections from the firewall
   * @return true if successful, false otherwise
   */
  public boolean loadTcpConns() {
    if (!Communicator.readTcpConns(this)) {
      if (Communicator.commError == Communicator.COMM_OK) {
        // retry once, server connection may be reestablished now
        if (!Communicator.readTcpConns(this))
          return false;
      }
      else
        return false;
    }
    return true;
  }

  /**
   * Show the host properties dialog. The drag and drop panel calls this
   * method when the user double clicks on the host icon in edit mode.
   * @param parent Parent frame (drag and drop panel)
   */
  public void execute(Frame parent) {
    HostProperties hp;
    if (objectID.length() > 0)
      hp = new HostProperties(parent, "Host "+objectID, this);
    else
      hp = new HostProperties(parent, "Host properties", this);
    hp.show();
  }

  /**
   * Show the firewall action window. The drag and drop panel calls this
   * method when the user double click on the host icon and edit mode
   * is diabled.
   * @param parent Parent frame (drag and drop panel)
   * @param g Graphics context of the drag and drop canvas
   * @param x_offset Horizontal offset of the virtual viewport
   * @param y_offset Vertical offset of the virtual viewport
   */
  public void userAction(Frame parent, Graphics g, int x_offset, int y_offset) {
    if (!isFirewall) {
      NoticeBox nb = new NoticeBox(parent, "This is not a firewall!");
      nb.show();
      return;
    }
    parent.setCursor(Frame.WAIT_CURSOR);
    if (!Communicator.openServer(HostAddresses.getFirstAddress())) {
      if (Communicator.commError == Communicator.COMM_NOCONN) {
        ErrorBox eb = new ErrorBox(parent, "Cannot connect to sfrelay!");
        eb.show();
        parent.setCursor(Frame.DEFAULT_CURSOR);
        return;
      }
      else {
        ErrorBox eb = new ErrorBox(parent, "Cannot connect to "+
          Communicator.resolveIP(HostAddresses.getFirstAddress()));
        eb.show();
        parent.setCursor(Frame.DEFAULT_CURSOR);
        return;
      }
    }
    if (actionFrame == null) {
      actionFrame = new FirewallAction(this);
      Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
      actionFrame.resize(screensize.width/6*5, screensize.height/4*3); 
    }
    actionFrame.show();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  /**
   * Generate firewall configuration file and write it to the firewall
   * @return true on success, false otherwise
   */
  public boolean generateConfig() {
    if (!isFirewall)
      return false;
    try {
      RelayOutputStream ros = new RelayOutputStream
        (HostAddresses.getFirstAddress(), Custom.GENERATE_CONFIG);
      PrintStream ps = new PrintStream(ros);

      ps.println("#");
      ps.println("# sf firewall configuration file");
      ps.println("# ------------------------------");
      ps.println("# automatically generated "+(new Date())+" for "+objectID);
      ps.println("# DO NOT EDIT THIS FILE!");
      ps.println("#");
      ps.println();

      ps.println("setup");
      ps.println();
      if (!internalAddresses.empty()) {
        ps.println("internalnets");
        internalAddresses.printAddressesFormatted(ps, 2, Macro.MACRO_FIRSTPORT, Macro.MACRO_LASTPORT);
        ps.println(";");
        ps.println();
      }
      ps.println("mail_default \""+mgDomain.defaultMail+"\";");
      ps.println();

      int ownidx = mgDomain.Firewalls.indexOf(this);
      Enumeration en = mgDomain.Rules.elements();
      Rule r;
      boolean hasRules = false;
      while (en.hasMoreElements()) {
        r = (Rule)en.nextElement();
        if (r.priority && r.validFor[ownidx]) {
          if (!hasRules) {
            hasRules = true;
            ps.println("priority_rules");
            ps.println();
          }
          r.printRule(ps, mgDomain, this);
        }
      }

      ps.println("rules");
      ps.println();
      en = mgDomain.Rules.elements();
      while (en.hasMoreElements()) {
        r = (Rule)en.nextElement();
        if (!r.priority && r.validFor[ownidx])
          r.printRule(ps, mgDomain, this);
      }

      ps.println("notification");
      ps.println();
      Notification.printAllLevels(ps, mgDomain.NLevels);

      ps.println("end.");

      ros.close();
      ps.close();

      return Communicator.checkConfig(HostAddresses.getFirstAddress());
    } // try
    catch (Exception e) {
      return false;
    }
  }

  /**
   * Reconfigure firewall. 
   * @return true on success, false otherwise
   */
  public boolean reconfig() {
    return Communicator.reconfig(HostAddresses.getFirstAddress());
  }

  // persistence methods
  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    super.write(ps);
    ps.writePersistent(",HostAddresses=", HostAddresses);
    ps.writePersistent(",internalAddresses=", internalAddresses);
    ps.writeBoolean(",", isFirewall);
    ps.writeBoolean(",", transparent);
    ps.writeBoolean(",", isServer);
    ps.writeBoolean(",", publicServer);
    ps.writeInt(",Templates=", hostTemplates.length);
    for (int i=0; i < hostTemplates.length; i++)
      ps.writeBoolean(",", hostTemplates[i]);
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    super.read(ps);
    HostAddresses = (AddrList)ps.readPersistent(",HostAddresses=");
    internalAddresses = (AddrList)ps.readPersistent(",internalAddresses=");
    isFirewall = ps.readBoolean(",");
    transparent = ps.readBoolean(",");
    isServer = ps.readBoolean(",");
    publicServer = ps.readBoolean(",");
    int tlen = ps.readInt(",Templates=");
    boolean btmp[] = new boolean[tlen];
    for (int i=0; i < tlen; i++)
      btmp[i] = ps.readBoolean(",");
    if (tlen == hostTemplates.length)
      hostTemplates = btmp;
  }

  // overriding Object methods

  /**
   * @return string describing the host
   */
  public String toString() {
    return "Host: "+objectInfo();
  }

  private static java.awt.image.ImageProducer bSource = null;

  // host properties
  protected AddrList HostAddresses = new AddrList();
  protected AddrList internalAddresses = null; // automatically calculated
  protected boolean isFirewall = false; // is this host a firewall?
  protected boolean transparent = true; // route traffic from outside to outside
  protected boolean isServer = false; // is this host a server (e.g.WWW) ?
  protected boolean publicServer = true; // is this server a public server, 
                                         // i.e. access from the internet allowed?
  protected boolean hostTemplates[];

  // variables for sfc show data
  protected transient boolean isLoaded = false;
  protected transient byte addr[][];
  protected transient byte mask[][];
  protected transient short port[];
  protected transient short prend[];
  protected transient Vector rules;
  protected transient String varnames[];
  protected transient int varvalues[][];
  protected transient byte varaddr[][][];
  protected transient Date vartimeout[][];
  protected transient Vector tcpconns;
  protected transient Date actDate; // server time and date when reading tcp connections

  private transient FirewallAction actionFrame = null;

} // Host

// Dialog to enter the host settings
class HostProperties extends Dialog {

  public HostProperties(Frame parent, String title, Host hst) {
    super(parent, title, true);
    host = hst;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(5,5,5,5);
    setLayout(gbl);

    // Name panel
    Utils.add_component(this, new ImageLabel(host.icon), gbl, gbc, 0, 0, 1, 1, 0, 0);
    Utils.add_component(this, new Label("Identifier: "), gbl, gbc, 1, 0, 1, 1, 0, 0);
    IDtext = new TextField(host.objectID);
    Utils.add_component(this, IDtext, gbl, gbc, 2, 0, 3, 1, 100, 0);

    // Address panel
    Panel ap = host.HostAddresses.editPanel();
    Utils.add_component(this, ap, gbl, gbc, 0, 1, 3, 8, 80, 100);

    // Host type panel
    FirewallBox = new Checkbox("Firewall");
    FirewallBox.setState(host.isFirewall);
    FirewallMore = new Button("More...");
    Utils.add_component(this, FirewallBox, gbl, gbc, 3, 2, 1, 1, 0, 0);
    Utils.add_component(this, FirewallMore, gbl, gbc, 4, 2, 1, 1, 0, 0);
    ServerBox = new Checkbox("Server");
    ServerBox.setState(host.isServer);
    ServerMore = new Button("More...");
    Utils.add_component(this, ServerBox, gbl, gbc, 3, 3, 1, 1, 0, 0);
    Utils.add_component(this, ServerMore, gbl, gbc, 4, 3, 1, 1, 0, 0);

    // Buttons
    OKbutton = new Button("OK");
    gbc.fill = GridBagConstraints.NONE;
    Utils.add_component(this, OKbutton, gbl, gbc, 0, 9, 5, 1, 0, 0);

    // Window size
    Graphics g = parent.getGraphics();
    FontMetrics fm = g.getFontMetrics();
    resize(fm.stringWidth("255.255.255.255/255.255.255.255   ") * 2,
           fm.getHeight() * 20 );
    g.dispose();
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target==OKbutton) {  
      host.objectID = IDtext.getText().trim();
      dispose();
      return true;
    }
    else if (evt.target==IDtext) {
      host.objectID = ((String)arg).trim();
      setTitle("Host "+host.objectID);
      return true;
    }
    else if (evt.target==FirewallBox) {
      host.isFirewall = FirewallBox.getState();
      return true;
    }
    else if (evt.target==ServerBox) {
      host.isServer = ServerBox.getState();
      return true;
    }
    else if (evt.target==FirewallMore) {
      if (host.isFirewall) {
        AutoconfSettings as = new AutoconfSettings(
          Utils.getParentFrame(this), "Firewall Settings", host, 
          Templates.startFirewall, Templates.numberFirewall);
        as.show();
      }
      return true;
    }
    else if (evt.target==ServerMore) {
      if (host.isServer) {
        AutoconfSettings as = new AutoconfSettings( 
          Utils.getParentFrame(this), "Server Settings", host, 
          Templates.startServer, Templates.numberServer);
        as.show();
      } 
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY && evt.target == this) {
      if (isEnabled())
        dispose();
      return true;
    }
    else
      return super.handleEvent(evt);
  }

  private Host host;
  private Button OKbutton;
  private TextField IDtext;
  private Checkbox FirewallBox;
  private Checkbox ServerBox;
  private Button FirewallMore;
  private Button ServerMore;
} // HostProperties

// Dialog to enter the firewall and server access profiles
class AutoconfSettings extends Dialog {

  public AutoconfSettings(Frame p, String title, Host hst, int st, int no) {
    super(p, title, true);
    start = st;
    number = no;
    host = hst;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(5,5,5,5);
    setLayout(gbl);

    // Checkboxes
    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.NORTHWEST;

    if (start == Templates.startServer) { // server dialog
      publicServerBox = new Checkbox("Public access to this server");
      publicServerBox.setState(host.publicServer);
      Utils.add_component(this, publicServerBox, gbl, gbc, 0, 0, 4, 1, 0, 0);
    }

    if (start == Templates.startFirewall) { // firewall dialog
      transparentBox = new Checkbox("Allow packets from outside to outside");
      transparentBox.setState(host.transparent);
      Utils.add_component(this, transparentBox, gbl, gbc, 0, 0, 4, 1, 0, 0);
    }

    CheckboxArray = new Checkbox[number];
    int half = (int)Math.floor(((double)number / 2.0) + 0.5);
    for (int i=start; i < start + number; i++) {
      CheckboxArray[i-start] = new Checkbox(Templates.prompt[i]);
      CheckboxArray[i-start].setState(host.hostTemplates[i-Templates.startServer]);
      if (i-start < half)
        Utils.add_component(this, CheckboxArray[i-start], gbl, gbc, 
                            0, i-start+1, 2, 1, 0, 0);
      else
        Utils.add_component(this, CheckboxArray[i-start], gbl, gbc, 
                            2, i-start-half+1, 2, 1, 0, 0);
    }

    // Buttons
    OKbutton = new Button("OK");
    Utils.add_component(this, OKbutton, gbl, gbc, 0, half+1, 4, 1, 0, 0);

    // Window size
    pack();
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target==OKbutton) {
      dispose();
      return true;
    }
    else if (evt.target == publicServerBox) {
      host.publicServer = publicServerBox.getState();
      return true;
    }
    else if (evt.target == transparentBox) {
      host.transparent = transparentBox.getState();
      return true;
    }
    else if (evt.target instanceof Checkbox) {
      for (int i=start; i < start+number; i++) 
        if (evt.target==CheckboxArray[i-start])
          host.hostTemplates[i-Templates.startServer] = CheckboxArray[i-start].getState();
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY && evt.target == this) {
      if (isEnabled())
        dispose();
      return true;
    }
    else
      return super.handleEvent(evt);
  }

  private Host host;
  int start;
  int number;

  private Checkbox publicServerBox;
  private Checkbox transparentBox;
  private Checkbox CheckboxArray[]; 
  private Button OKbutton;

} // AutoconfSettings

// HostIcon
class cpu1ImageSource extends java.awt.image.MemoryImageSource {

  final static int bWidth = 23;
  final static int bHeight = 25;
  private static int[] bPixels = { // color model is AARRGGBB

    0x00ffffff, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0x00ffffff,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xffbfbfbf, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 0xff0000fe, 
    0xff0000fe, 0xff0000fe, 0xff0000fe, 0xffbfbfbf, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xffbfbfbf,  

    0xffbfbfbf, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xffbfbfbf,  

    0x00ffffff, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0x00ffffff,  

    0x00ffffff, 
    0x00ffffff, 0x00ffffff, 0xffbfbfbf, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xffbfbfbf, 0x00ffffff, 
    0x00ffffff, 0x00ffffff,  

    0x00ffffff, 
    0x00ffffff, 0xffbfbfbf, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 
    0xfffefe00, 0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 
    0xfffefefe, 0xfffefe00, 0xfffefefe, 0xfffefe00, 0xffbfbfbf, 
    0x00ffffff, 0x00ffffff,  

    0x00ffffff, 
    0x00ffffff, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 0xffbfbfbf, 
    0x00ffffff, 0x00ffffff,  

    0x00ffffff, 
    0x00ffffff, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 
    0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 
    0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 
    0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 
    0x00ffffff, 0x00ffffff,  

    0x00ffffff, 
    0x00ffffff, 0xffbfbfbf, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 
    0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 
    0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 
    0xfffefe00, 0xffbfbfbf, 0xfffefe00, 0xffbfbfbf, 0xffbfbfbf, 
    0x00ffffff, 0x00ffffff,  
    };

  public cpu1ImageSource() {
    super( bWidth, bHeight, bPixels, 0, bWidth );
  }
}
