
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.io.*;
import sfclasses.*;

/**
 * This class implements an output stream to write data to a local file
 * via the relay. It uses the I/O functions implemented in the Communicator
 * class. Filter streams can be used on top of this stream to enhance
 * functionality.
 * @version 1.0 26 Jan 1997
 * @author Roland E. Schmid
 */

public class LocalOutputStream extends OutputStream {

  /**
   * Initialize the stream and open the file for writing
   * @param filename Name of the file including the full path
   */
  public LocalOutputStream(String filename) {
    closed = !Communicator.writeLocalFile(filename);
  }

  /**
   * Write one byte to the file. The output is buffered and will
   * be written when flush() or close() is called.
   * @param b Byte to write
   */
  public void write(int b) throws IOException {
    if (closed)
      throw new IOException();
    buffer[bufptr] = (byte)b;
    bufptr++;
    if (bufptr == bufsize)
      flush();
  }

  /**
   * Write the buffer to the file.
   */
  public void flush() throws IOException {
    if (closed)
      throw new IOException();
    if (bufptr == 0)
      return;
    try {
      Communicator.writeHeader(Communicator.CLI_FILE_DATA, bufptr);
      Communicator.out.write(buffer, 0, bufptr);
      bufptr = 0;
      Communicator.readHeader();
      if ((Communicator.readLength != 0) || (Communicator.readCommand != Communicator.CLI_OK))
        throw new IOException();
    }
    catch (Exception e) {
      throw new IOException();
    }
  }

  /**
   * Close the file and the output stream.
   */
  public void close() throws IOException {
    if (closed)
      return;
    if (bufptr > 0)
      flush();
    Communicator.writeHeader(Communicator.CLI_FILE_DATA, 0);
    bufptr = 0;
    closed = true;
    try {
      // read OK for FILE_DATA
      Communicator.readHeader();
      if ((Communicator.readLength != 0) || (Communicator.readCommand != Communicator.CLI_OK))
        throw new IOException();
    }
    catch (Exception e) {
      throw new IOException();
    }
  }

  /**
   * Check if the output stream is closed. 
   * @return true if the stream is closed and false if data can be written.
   */
  public boolean isClosed() {
    return closed;
  }

  private static final int bufsize = 1000;
  private boolean closed;
  private byte[] buffer = new byte[bufsize];
  private int bufptr = 0;

} // LocalOutputStream

