
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.awt.*;
import java.net.*;
import java.applet.*;
import java.util.*;
import java.io.*;
import Acme.Widgets.*;

/**
 * <b>Toplevel Frame of Remote Config Tool</b><br>
 * This class implements the main window of the control panel. The frame
 * is constructed when the user clicks on a button in the start page (see
 * sfRemote.java). It contains a drag and drop panel for the network
 * topology in the client area and provides access to all global 
 * configuration functions.
 * @version 1.0 12 Dec 1996
 * @author Roland E. Schmid
 */

public class ManageControl extends Frame {

  /**
   * Initialize new main window<br>
   * The window is set up and a new ManageDomain object is created
   * holding the configuration data.
   * @param titleString Initial window title
   * @ac Applet context, used for displaying strings in the status line
   * of the web browser and to open new browser windows.
   */
  public ManageControl(String titleString, AppletContext ac) {

    // window title bar
    if (titleString != null)
      setTitle(titleString);

    // create menus
    mbar = new MenuBar();

    FileMenu = new Menu("File");
    FileMenu.add(new MenuItem("Load from Firewall"));
    FileMenu.add(new MenuItem("Save to Firewalls"));
    FileMenu.add(new MenuItem("Load from Local File"));
    FileMenu.add(new MenuItem("Save to Local File"));
    FileMenu.add(new MenuItem("Close"));
    FileMenu.addSeparator();
    FileMenu.add(new MenuItem("Quit"));
    mbar.add(FileMenu);

    EditMenu = new Menu("Edit"); 
    EditMenu.add(new MenuItem("General"));
    EditMenu.add(new MenuItem("Autoconfig"));
    EditMenu.add(new MenuItem("Rules"));
    EditMenu.add(new MenuItem("Notification Levels"));
    EditMenu.add(new MenuItem("Macros"));
    EditMenu.addSeparator();
    CheckboxMenuItem cmi = new CheckboxMenuItem("Edit Topology");
    cmi.setState(topologyEdit);
    EditMenu.add(cmi);
    mbar.add(EditMenu);

    ReportMenu = new Menu("Reports");
    ReportMenu.add(new MenuItem("Configuration Report"));
    mbar.add(ReportMenu);

    HelpMenu = new Menu("Help");
    HelpMenu.add(new MenuItem("User's Guide"));
    HelpMenu.add(new MenuItem("About"));
    mbar.add(HelpMenu);
    mbar.setHelpMenu(HelpMenu);
    setMenuBar(mbar);

    // initialize data
    appletContext = ac;
    if (titleString != null) {
      byte host[] = Communicator.resolve(titleString);
      RelayInputStream ris = 
        new RelayInputStream(host, Custom.CONF_DIR + Custom.GLOBAL_CONFIG);
      if (ris.isEOF())
        mgDomain = null;
      else {
        PersistentInputStream is = new PersistentInputStream(ris);
        mgDomain = ManageDomain.loadDomainConfig(is);
      }
    }
    else
      mgDomain = new ManageDomain();
    if (mgDomain == null) { // host is unreachable
      prepareQuit();
      return; 
    }
    // create drag and drop panel
    try {
      hostClass = Class.forName("sfclasses.Host");
      netClass = Class.forName("sfclasses.Net");
      internetClass = Class.forName("sfclasses.Internet");
    }
    catch (ClassNotFoundException e) {
      System.out.println("Fatal error: class Host, Net or Internet not found!");
      terminated = true;
      return;
    }
    drawClass = hostClass;

    // create panels
    ddPanel = new DragDrop(mgDomain.getTopology(), drawClass, topologyEdit);
    Toolbar = new Panel();
    Toolbar.setLayout(new FlowLayout(FlowLayout.LEFT));
    CheckboxGroup g = new CheckboxGroup();
    Toolbar.add(hostBox = new Checkbox("Host", g, true));
    Toolbar.add(netBox = new Checkbox("Net", g, false));
    Toolbar.add(internetBox = new Checkbox("Internet", g, false));
    setLayout(new BorderLayout());
    add("Center", ddPanel);

    // set window title
    if (mgDomain.DomainName.length() > 0)
      setTitle(mgDomain.DomainName);
    else
      setTitle("Unnamed domain");
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target instanceof MenuItem) {
      if (arg.equals("Quit")) {
        prepareQuit();
        return true;
      }
      else if (arg.equals("Save to Firewalls")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        SaveData sd = new SaveData(mgDomain, this, false, false);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        sd.start();
        return true;
      }
      else if (arg.equals("Save to Local File")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        SaveData sd = new SaveData(mgDomain, this, false, true);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        sd.start();
        return true;
      }
      else if (arg.equals("Load from Local File")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        LoadData ld = new LoadData(mgDomain, this, LoadData.LOCAL);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        ld.start();
        return true;
      }
      else if (arg.equals("Load from Firewall")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        LoadData ld = new LoadData(mgDomain, this, LoadData.REMOTE);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        ld.start();
        return true;
      }
      else if (arg.equals("Close")) {
        LoadData ld = new LoadData(mgDomain, this, LoadData.NONE);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        ld.start();
        return true;
      }
      else if (arg.equals("General")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing rules!");
          eb.show();
          return true;
        }
        mgDomain.changed = true;
        mgDomain.autoconfValid = false;
        GeneralSettings gs;
        if (mgDomain.DomainName.length() == 0)
          gs = new GeneralSettings(this, "General Settings", mgDomain);
        else
          gs = new GeneralSettings(this, mgDomain.DomainName, mgDomain);
        gs.show();
        return true;
      }
      else if (arg.equals("Rules")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        mgDomain.changed = true;
        if (editRuleFrame == null) {
          editRuleFrame = new RuleEditor(mgDomain);
          Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
          editRuleFrame.resize(screensize.width/6*5, screensize.height/4*3);
          editRuleFrame.show();
        }
        else
          editRuleFrame.showAgain();
        if (!mgDomain.autoconfValid) {
          WarningBox wb = new WarningBox(editRuleFrame, "Autoconfig not executed. Rules may not be valid.");
          wb.show();
        }
        return true;
      }
      else if (arg.equals("Autoconfig")) {
        if (topologyEdit) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing the topology!");
          eb.show();
          return true;
        }
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing rules!");
          eb.show();
          return true;
        }
        mgDomain.changed = true;
        AutoconfThread at = new AutoconfThread(mgDomain, this);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        at.start();
        return true;
      }
      else if (arg.equals("Notification Levels")) {
        NotificationSelect notificationSelect = new NotificationSelect(mgDomain, this);
        notificationSelect.show();
        return true;
      }
      else if (arg.equals("Macros")) {
        MacroSelect macroSelect = new MacroSelect(mgDomain, this);
        macroSelect.show();
        return true;
      }
      else if (arg.equals("Edit Topology")) {
        if (editRuleFrame != null && editRuleFrame.isShowing()) {
          ErrorBox eb = new ErrorBox(this, "Function not available while editing rules!");
          eb.show();
          topologyEdit = false;
          ((CheckboxMenuItem)evt.target).setState(topologyEdit);
          return true;
        }
        topologyEdit = ((CheckboxMenuItem)evt.target).getState();
        mgDomain.changed = true;
        mgDomain.autoconfValid = false;
        if (topologyEdit) {
          ddPanel.enableEdit();
          add("North", Toolbar);
          validate();
        }
        else {
          if (!ddPanel.disableEdit()) {
            topologyEdit = true;
            ((CheckboxMenuItem)evt.target).setState(topologyEdit);
          }
          remove(Toolbar);
          mgDomain.topologyChanged();
          validate();
          NoticeBox nb = new NoticeBox(this, "Run autoconfiguration now to update the rules!");
          nb.show();
        }
        return true;
      }
      else if (arg.equals("Configuration Report")) {
        ReportThread rt = new ReportThread(mgDomain, this, ReportThread.REPORT_CONFIG);
        disable();
        setCursor(Frame.WAIT_CURSOR);
        rt.start();
        return true;
      }
      else if (arg.equals("User's Guide")) {
        try {
          URL u = new URL(Utils.fileURLPrefix()+Custom.DOC_DIR+"control.htm");
          appletContext.showDocument(u,"_blank");
        }
        catch (MalformedURLException e) {
          System.out.println("Malformed URL! "+e);
        }
        catch (Exception e) {
          System.out.println("Unexpected exception "+e);
        }
        return true;
      }
      else if (arg.equals("About")) {
        AboutDialog ab = new AboutDialog(this);
        ab.show();
        return true;
      }
      else
        return super.action(evt, arg);
    }
    else if (evt.target == hostBox) {
      drawClass = hostClass;
      ddPanel.setClass(drawClass);
      return true;
    }
    else if (evt.target == netBox) {
      drawClass = netClass;
      ddPanel.setClass(drawClass);
      return true;
    }
    else if (evt.target == internetBox) {
      drawClass = internetClass;
      ddPanel.setClass(drawClass);
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY) {
      if (isEnabled())
        prepareQuit();
      return true;
    }
    else
      return super.handleEvent(evt);
  }

  protected void prepareQuit() {
    if (mgDomain != null && mgDomain.changed) {
      SaveData sd = new SaveData(mgDomain, this, true, false);
      disable();
      setCursor(Frame.WAIT_CURSOR);
      sd.start();
      return;
    }     

    /* cleanup stuff */
    Communicator.dispose();
    hide();
    ddPanel = null;
    mgDomain = null;
    terminated = true;
  }

  protected void setManageDomain(ManageDomain newmd) {
    remove(ddPanel);
    ddPanel = new DragDrop(newmd.getTopology(), drawClass, topologyEdit);
    add("Center", ddPanel);
    validate();

    if (newmd.DomainName.length() > 0)
      setTitle(newmd.DomainName);
    else
      setTitle("Unnamed domain");

    mgDomain = newmd;
  }

  // public data
  public boolean terminated = false;
  public static AppletContext appletContext;

  // private data

  // gui data
  private DragDrop ddPanel;
  private MenuBar mbar;
  private Menu FileMenu;
  private Menu EditMenu;
  private Menu ReportMenu;
  private Menu HelpMenu;
  private Panel Toolbar;
  private Checkbox hostBox;
  private Checkbox netBox;
  private Checkbox internetBox;
  private RuleEditor editRuleFrame;

  // application data
  private ManageDomain mgDomain;
  private boolean topologyEdit = false;
  private Class drawClass, hostClass, netClass, internetClass;

} // ManageControl

// Thread used for saving data. A separate thread is necessary, because
// it is not possible to wait for the user response when showing a dialog
// box from the event handling loop.
class SaveData extends Thread {

  public SaveData(ManageDomain md, ManageControl p, boolean ask, boolean sl) {
    mgDomain = md;
    parent = p;
    askDialog = ask;
    saveLocal = ask ? false : sl;
  }

  public void run() {

    if (askDialog) {
      YesNoCancelBox ync = new YesNoCancelBox(parent, "Data not saved", "Save data to firewalls before exiting?");
      ync.show();
      parent.setCursor(Frame.DEFAULT_CURSOR);
      switch (ync.getAnswer()) {
        case YesNoCancelBox.YES: // save and quit
          break;
        case YesNoCancelBox.NO: // don't save, quit
          mgDomain.changed = false;
          parent.prepareQuit();
          return;
        case YesNoCancelBox.CANCEL: // don't save, don't quit
          parent.enable();
          return;
      }
      parent.setCursor(Frame.WAIT_CURSOR);
    }

    boolean success = true;
    if (!saveLocal) { // save data to firewalls
      BusyBox b = new BusyBox(parent, "Save in progress. Please wait!");
      b.show(); // show busy box, returns immediately
      Enumeration en = mgDomain.Firewalls.elements();
      Host f;
      byte faddr[];
      RelayOutputStream ros;
      PersistentOutputStream os;
      while (en.hasMoreElements()) {
        f = (Host)en.nextElement();
        faddr = f.HostAddresses.getFirstAddress();
        try {
          ros = new RelayOutputStream(faddr, Custom.GLOBAL_CONFIG);
          if (ros.isClosed())
            throw new IOException();
          os = new PersistentOutputStream(ros);
          if (!mgDomain.saveDomainConfig(os))
            throw new IOException();
          ros.close();
          os.close();
        } // try
        catch (IOException e) {
          ErrorBox eb = new ErrorBox(parent, 
            "Error saving data to "+f.objectID+" ("+Utils.printIP(faddr)+").");
          eb.show();
          success = false;
        } // catch
      } // while
      b.done();
    } // if
    else { // save data to local file
      StringDialog strd = new StringDialog(parent, "Filename:");
      strd.show();
      if (strd.getAnswer() != StringDialog.OK) {
        parent.enable();
        parent.setCursor(Frame.DEFAULT_CURSOR);
        return;
      }
      String filename = strd.getString();
      BusyBox b = new BusyBox(parent, "Save in progress. Please wait!");
      b.show(); // show busy box, returns immediately
      try {
        LocalOutputStream los = new LocalOutputStream(filename);
        if (los.isClosed())
          throw new IOException("Error opening file");
        PersistentOutputStream os = new PersistentOutputStream(los);
        if (!mgDomain.saveDomainConfig(os))
          throw new IOException("Error writing data");
        los.close();
        os.close();
        b.done();
      } // try
      catch (IOException e) {
        b.done();
        ErrorBox eb = new ErrorBox(parent, 
          "Error saving data to "+filename+". Data not saved! ("+Communicator.commError+")");
        eb.show();
        success = false;
      } // catch
    }

    if (success) {
      mgDomain.changed = false;
      NoticeBox nb = new NoticeBox(parent, "Data saved successfully!");
      nb.show();
      if (askDialog) // quit after saving
        parent.prepareQuit();
    }

    parent.enable();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  private ManageDomain mgDomain;
  private ManageControl parent; // parent frame
  private boolean askDialog, saveLocal;

} // SaveData 

// This thread is used for loading data
class LoadData extends Thread {

  public LoadData(ManageDomain md, ManageControl p, int l) {
    mgDomain = md;
    parent = p;
    load = l;
  }

  public void run() {

    if (mgDomain.changed) {
      YesNoCancelBox ync = new YesNoCancelBox(parent, "Data not saved", "Save data to firewalls?");
      ync.show();
      switch (ync.getAnswer()) {
        case YesNoCancelBox.YES: // save and load
          SaveData sd = new SaveData(mgDomain, parent, false, false);
          sd.start();
          while (sd.isAlive()) {
            try {
              sd.join(); // wait for save thread to complete
              break;
            }
            catch (Exception ignore) {}
          }
          break;
        case YesNoCancelBox.NO: // don't save, load
          break;
        case YesNoCancelBox.CANCEL: // don't save, abort load
          parent.enable();
          parent.setCursor(Frame.DEFAULT_CURSOR);
          return;
      }
    }

    if (load == NONE) 
      parent.setManageDomain(new ManageDomain());

    if (load == REMOTE) {
      StringDialog strd = new StringDialog(parent, "Hostname or IP address:");
      strd.show();
      if (strd.getAnswer() != StringDialog.OK) {
        parent.enable();
        parent.setCursor(Frame.DEFAULT_CURSOR);
        return;
      }
      String hostname = strd.getString();
      byte host[] = Communicator.resolve(hostname);
      if (host == null) {
        ErrorBox eb;
        switch (Communicator.commError) {
          case Communicator.COMM_NOCONN:
            eb = new ErrorBox(parent, "Cannot connect to sfrelay!");
            break;
          case Communicator.COMM_RESOLVE:
            eb = new ErrorBox(parent, "Unknown host "+hostname);
            break;
          default:
            eb = new ErrorBox(parent, "Error while checking hostname. ("+Communicator.commError+")");
            break;
        }
        eb.show();
      }
      else {
        ManageDomain newmd;
        RelayInputStream ris =
          new RelayInputStream(host, Custom.CONF_DIR + Custom.GLOBAL_CONFIG);
        if (ris.isEOF())
          newmd = null;
        else {
          PersistentInputStream is = new PersistentInputStream(ris);
          newmd = ManageDomain.loadDomainConfig(is);
        }

        if (newmd == null) { // error reading file
          ErrorBox eb = new ErrorBox(parent, "Error loading data! ("+Communicator.commError+")");
          eb.show();
        }
        else
          parent.setManageDomain(newmd);
      }
    }

    if (load == LOCAL) {
      StringDialog strd = new StringDialog(parent, "Filename:");
      strd.show();
      if (strd.getAnswer() != StringDialog.OK) {
        parent.enable();
        parent.setCursor(Frame.DEFAULT_CURSOR);
        return;
      }
      String filename = strd.getString();
    
      LocalInputStream lis = new LocalInputStream(filename);
      ManageDomain newmd;
      if (lis.isEOF())
        newmd = null;
      else {
        PersistentInputStream is = new PersistentInputStream(lis);
        newmd = ManageDomain.loadDomainConfig(is);
      }
      if (newmd == null) { // error reading file
        ErrorBox eb = new ErrorBox(parent, "Error loading data! ("+Communicator.commError+")");
        eb.show();
      }
      else
        parent.setManageDomain(newmd);
    }

    parent.enable();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  private ManageDomain mgDomain;
  private ManageControl parent; // parent frame
  private int load; // if false create empty domain

  public static final int NONE = 0;
  public static final int LOCAL = 1;
  public static final int REMOTE = 2;

} // LoadData

// This thread executes autoconfiguration
class AutoconfThread extends Thread {

  public AutoconfThread(ManageDomain md, Frame p ) {
    mgDomain = md;
    parent = p;
  }

  public void run() {
    BusyBox bb = new BusyBox(parent, "Running Autoconfiguration. Please wait!");
    bb.show();
    try {
      sleep(500); // time to show busy box
    }
    catch (Exception e) {}
    mgDomain.autoConfig();
    bb.done();
    NoticeBox nb = new NoticeBox(parent, "Autoconfiguration completed successfully!");
    nb.show();
    parent.enable();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  private ManageDomain mgDomain;
  private Frame parent;
} // AutoconfThread

// This thread generates and shows reports
class ReportThread extends Thread {

  public ReportThread(ManageDomain md, Frame p, int rt) {
    mgDomain = md;
    parent = p;
    reportType = rt;
  }

  public void run() {
    BusyBox bb = new BusyBox(parent, "Generating report. Please wait!");
    bb.show();
    String filename = Communicator.getTempName();
    if (filename == null) {
      ErrorBox eb = new ErrorBox(parent, "Cannot create temporary file name! ("+Communicator.commError+")");
      bb.done();
      eb.show();
      parent.enable();
      parent.setCursor(Frame.DEFAULT_CURSOR);
      return;
    }
 
    try {
      LocalOutputStream los = new LocalOutputStream(filename);
      if (los.isClosed()) 
        throw new IOException();
      PrintStream ps = new PrintStream(los);

      switch (reportType) {
        case REPORT_CONFIG:
          writeConfigReport(ps);
          break;
        default:
          ErrorBox eb = new ErrorBox(parent, "Unknown report type!");
          bb.done();
          eb.show();
          parent.enable();
          parent.setCursor(Frame.DEFAULT_CURSOR);
          return;
      }
      los.close();
      ps.close();

      URL u = new URL(Utils.fileURLPrefix()+filename);
      ManageControl.appletContext.showDocument(u,"_blank");
    }
    catch (IOException e) {
      ErrorBox eb = new ErrorBox(parent, "Error writing file! ("+Communicator.commError+")");
      eb.show();
    }
    finally {
      bb.done();
    }

    parent.setCursor(Frame.DEFAULT_CURSOR);
    parent.enable();
  }

  /**
   * generate configuration report
   */
  private void writeConfigReport(PrintStream ps) {
    ps.println("<html><head><title>sf Configuration Report</title></head><body>");
    ps.println("<h1>sf Configuration Report</h1>");
    ps.println("<p>Report generated "+(new Date()).toString());
    if (mgDomain.DomainName.length() > 0)
      ps.println("<h2>Management domain <i>"+mgDomain.DomainName+"</i></h2>");

    Enumeration en = mgDomain.getTopology().getAllVertices();
    ManageObject mo;
    Host h;
    Net n;
    Internet in;
    while (en.hasMoreElements()) {
      mo = (ManageObject)en.nextElement();
      if (mo instanceof Host) {
        h = (Host)mo;
        ps.println("<h3><a name="+mo.identString()+">Host <i>"+h.objectID+"</i></a></h3>");
        ps.println("<ul>");
        if (h.level == 1)
          ps.println("<li>Level: "+h.level+" firewall hop to nearest Internet connection");
        else if (h.level >= 0)
          ps.println("<li>Level: "+h.level+" firewall hops to nearest Internet connection");
        else
          ps.println("<li>No Internet connection");
        if (h.isFirewall)
          if (h.transparent)
            ps.println("<li>Firewall (allows packets from outside to outside)");
          else
            ps.println("<li>Firewall (blocks packets from outside to outside)");
        if (h.isServer && h.publicServer)
          ps.println("<li>Server (public access)");
        if (h.isServer && !h.publicServer)
          ps.println("<li>Server (internal access only)");
        ps.println("<li>Addresses:");
        writeAddresses(ps, h.HostAddresses);
        if (h.isFirewall) {
          ps.println("<li>Internalnets:");
          writeAddresses(ps, h.internalAddresses);
        }
        writeNeighbors(ps, mo);
        ps.println("</ul>");
      }
      else if (mo instanceof Net) {
        n = (Net)mo;
        ps.println("<h3><a name="+mo.identString()+">Network <i>"+n.objectID+"</i></a></h3>");
        ps.println("<ul>");
        if (n.level == 1)
          ps.println("<li>Level: "+n.level+" firewall hop to nearest Internet connection");
        else if (n.level >= 0)
          ps.println("<li>Level: "+n.level+" firewall hops to nearest Internet connection");
        else
          ps.println("<li>No Internet connection");
        ps.println("<li>Addresses:");
        writeAddresses(ps, n.NetAddresses);
        writeNeighbors(ps, mo);
        ps.println("</ul>");
      }
      else if (mo instanceof Internet) {
        in = (Internet)mo;
      }
      else {
        ps.println("<h3><a name="+mo.identString()+">Unknown object: <i>"+mo+"</i></a></h3>");
      }
    }

    ps.println("</body></html>");
  }

  private void writeAddresses(PrintStream ps, AddrList al) {
    if (al==null)
      return;
    Enumeration en = al.getAddresses().elements();
    AddrObj ao;
    ps.println("<ul>");
    while (en.hasMoreElements()) {
      ao = (AddrObj)en.nextElement();
      ps.println("<li><tt>"+ao.toString()+"</tt>");
    }
    ps.println("</ul>");
  }

  private void writeNeighbors(PrintStream ps, ManageObject mo) {
    Enumeration en = mgDomain.getTopology().findNeighbors(mo);
    ManageObject n;
    ps.println("<li>Connected to:<ul>");
    while (en.hasMoreElements()) {
      n = (ManageObject)en.nextElement();
      if (n.objectID.length() > 0)
        if (n instanceof Internet)
          ps.println("<li>"+n.objectID);
        else
          ps.println("<li><a href=\"#"+n.identString()+"\">"+n.objectID+"</a>");
      else if (n instanceof Host)
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed host"+"</a>");
      else if (n instanceof Net)
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed network"+"</a>");
      else
        ps.println("<li><a href=\"#"+n.identString()+"\">Unnamed object"+"</a>");
    }
    ps.println("</ul>");
  }

  public static final int REPORT_CONFIG = 1; // configuration report

  private ManageDomain mgDomain;
  private Frame parent;
  private BusyBox busybox;
  private int reportType;
} // ReportThread

// This dialog box is used to enter the general access profile
// and the configuration client addresses
class GeneralSettings extends Dialog {

  public GeneralSettings(Frame p, String title, ManageDomain md) {
    super(p, title, true);
    parent = p;
    mgDomain = md;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(5,5,5,5);
    setLayout(gbl);

    // Name panel
    Utils.add_component(this, new Label("Name: "), gbl, gbc, 0, 0, 1, 1, 0, 0);
    DNameText = new TextField(mgDomain.DomainName);
    Utils.add_component(this, DNameText, gbl, gbc, 1, 0, 3, 1, 100, 0);

    // Default mail addresses
    Utils.add_component(this, new Label("Default mail:"), gbl, gbc, 0, 1, 1, 1, 0, 0);
    mailText = new TextField(mgDomain.defaultMail);
    Utils.add_component(this, mailText, gbl, gbc, 1, 1, 3, 1, 100, 0);

    int half = (int)Math.floor(((double)Templates.numberGeneral / 2.0) + 0.5);

    // Address Panel (configuration clients)
    addrPanel = mgDomain.configClients.editPanel();
    Utils.add_component(this, new Label("Configuration clients:"), gbl, gbc, 0, 2, 2, 1, 0, 0);
    Utils.add_component(this, addrPanel, gbl, gbc, 0, 3, 2, half-1, 100, 100);

    // Checkboxes
    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.NORTHWEST;
    CheckboxArray = new Checkbox[Templates.numberGeneral];
    for (int i=0; i < Templates.numberGeneral; i++) {
      CheckboxArray[i] = new Checkbox(Templates.prompt[i]);
      CheckboxArray[i].setState(mgDomain.generalTemplates[i]);
      if (i < half)
        Utils.add_component(this, CheckboxArray[i], gbl, gbc, 
                            2, i+2, 1, 1, 0, 0);
      else
        Utils.add_component(this, CheckboxArray[i], gbl, gbc, 
                            3, i+2-half, 1, 1, 0, 0);
    }

    // Buttons
    OKbutton = new Button("OK");
    Utils.add_component(this, OKbutton, gbl, gbc, 0, 2+half, 4, 1, 0, 0);

    // Window size
    pack();
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target==DNameText) {
      mgDomain.DomainName = ((String)arg).trim();
      setTitle(mgDomain.DomainName);
      parent.setTitle(mgDomain.DomainName);
      return true;
    }
    else if (evt.target==OKbutton) {
      mgDomain.DomainName = DNameText.getText().trim();
      if (mgDomain.DomainName.length() > 0)
        parent.setTitle(mgDomain.DomainName);
      dispose();
      return true;
    }
    else if (evt.target instanceof Checkbox) {
      for (int i=0; i < Templates.numberGeneral; i++) 
        if (evt.target==CheckboxArray[i])
          mgDomain.generalTemplates[i] = CheckboxArray[i].getState();
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY && evt.target == this) {
      if (isEnabled())
        dispose();
      return true;
    }
    else
      return super.handleEvent(evt);
  }

  private ManageDomain mgDomain;

  private Frame parent;

  private TextField DNameText;
  private TextField mailText;
  private Panel addrPanel;
  private Checkbox CheckboxArray[]; 
  private Button OKbutton;

} // GeneralSettings

// This dialog box provides a selection list for editing notification levels
class NotificationSelect extends Dialog implements Refreshable {

  public NotificationSelect(ManageDomain md, ManageControl p) {
    super(p, "Notification Levels", true);

    p.setCursor(Frame.WAIT_CURSOR);

    mgDomain = md;
    parent = p;

    gbl = new GridBagLayout();
    gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    nList = new List(15, false);
    Notification.constructList(nList, mgDomain.NLevels);
    OKbutton = new Button("OK");
    AddButton = new Button("Add new level");
    DeleteButton = new Button("Delete");
    ChangeButton = new Button("Change");

    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);

    Utils.add_component(this, OKbutton, gbl, gbc, 3, 0, 1, 1, 0, 0);
    Utils.add_component(this, AddButton, gbl, gbc, 3, 1, 1, 1, 0, 0);
    Utils.add_component(this, DeleteButton, gbl, gbc, 3, 2, 1, 1, 0, 0);
    Utils.add_component(this, ChangeButton, gbl, gbc, 3, 3, 1, 1, 0, 0);
    Utils.add_component(this, new Spacer(1,1), gbl, gbc, 3, 4, 1, 1, 100, 100);

    pack();

    p.setCursor(Frame.DEFAULT_CURSOR);
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target == OKbutton) {
      dispose();
      return true;
    }
    else if (evt.target == AddButton) {
      Notification ln;
      try {
        ln = (Notification)mgDomain.NLevels.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Notification not = new Notification
        (ln == null ? Notification.firstDynamic : ln.levelNumber + 1);
      mgDomain.NLevels.addElement(not);
      not.edit(this, parent, mgDomain, true);
      return true;
    }
    else if (evt.target == DeleteButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return true;
      Notification not = Notification.getLevel(str, mgDomain.NLevels);
      if (not == null)
        return true;
      if (not.levelNumber < Notification.firstDynamic) {
        ErrorBox eb = new ErrorBox(parent, "Cannot delete static notification level!");
        eb.show();
        return true;
      }
// !!! this doesn't work, cause we are in the event loop
//      OkCancelBox oc = new OkCancelBox(parent, "Deleting notification level "+str+".");
//      oc.show();
//      if (oc.getAnswer() != OkCancelBox.OK)
//        return true;
      mgDomain.NLevels.removeElement(not);
      refresh();
      return true;
    }
    else if (evt.target == ChangeButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return true;
      Notification not = Notification.getLevel(str, mgDomain.NLevels);
      if (not == null)
        return true;
      not.edit(this, parent, mgDomain, false);
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public void refresh() {
    parent.setCursor(Frame.WAIT_CURSOR);
    remove(nList);
    nList = new List(15, false);
    Notification.constructList(nList, mgDomain.NLevels);
    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);
    validate();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  private ManageDomain mgDomain;
  private ManageControl parent;
 
  private GridBagLayout gbl;
  private GridBagConstraints gbc;

  private List nList;
  private Button OKbutton;
  private Button AddButton;
  private Button DeleteButton;
  private Button ChangeButton;

} // NotificationSelect

// This dialog box implements a selection list for editing macros
class MacroSelect extends Dialog implements Refreshable {

  public MacroSelect(ManageDomain md, ManageControl p) {
    super(p, "Address and Service Macros", true);

    p.setCursor(Frame.WAIT_CURSOR);

    mgDomain = md;
    parent = p;

    gbl = new GridBagLayout();
    gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    nList = new List(15, false);
    Macro.constructList(nList, mgDomain.Macros);
    OKbutton = new Button("OK");
    AddButton = new Button("Add new macro");
    DeleteButton = new Button("Delete");
    ChangeButton = new Button("Change");

    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);

    Utils.add_component(this, OKbutton, gbl, gbc, 3, 0, 1, 1, 0, 0);
    Utils.add_component(this, AddButton, gbl, gbc, 3, 1, 1, 1, 0, 0);
    Utils.add_component(this, DeleteButton, gbl, gbc, 3, 2, 1, 1, 0, 0);
    Utils.add_component(this, ChangeButton, gbl, gbc, 3, 3, 1, 1, 0, 0);
    Utils.add_component(this, new Spacer(1,1), gbl, gbc, 3, 4, 1, 1, 100, 100);

    pack();

    p.setCursor(Frame.DEFAULT_CURSOR);
  }

  public boolean action(Event evt, Object arg) {
    if (evt.target == OKbutton) {
      dispose();
      return true;
    }
    else if (evt.target == AddButton) {
      Macro ln;
      try {
        ln = (Macro)mgDomain.Macros.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Macro m = new Macro
        (ln == null ? Macro.firstDynamic : ln.macroNumber - 1);
      mgDomain.Macros.addElement(m);
      m.edit(this, parent, mgDomain, true);
      return true;
    }
    else if (evt.target == DeleteButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return true;
      Macro m = Macro.getMacro(str, mgDomain.Macros);
      if (m == null)
        return true;
      if (m.macroNumber > Macro.firstDynamic) {
        ErrorBox eb = new ErrorBox(parent, "Cannot delete static macro!");
        eb.show();
        return true;
      }
// !!! this doesn't work, cause we are in the event loop
//      OkCancelBox oc = new OkCancelBox(parent, "Deleting macro "+str+".");
//      oc.show();
//      if (oc.getAnswer() != OkCancelBox.OK)
//        return true;
      mgDomain.Macros.removeElement(m);
      refresh();
      return true;
    }
    else if (evt.target == ChangeButton) {
      String str = nList.getSelectedItem();
      if (str == null)
        return true;
      Macro m = Macro.getMacro(str, mgDomain.Macros);
      if (m == null)
        return true;
      m.edit(this, parent, mgDomain, false);
      return true;
    }
    else
      return super.action(evt, arg);
  }

  public void refresh() {
    parent.setCursor(Frame.WAIT_CURSOR);
    remove(nList);
    nList = new List(15, false);
    Macro.constructList(nList, mgDomain.Macros);
    Utils.add_component(this, nList, gbl, gbc, 0, 0, 3, 15, 100, 100);
    validate();
    parent.setCursor(Frame.DEFAULT_CURSOR);
  }

  private ManageDomain mgDomain;
  private ManageControl parent;
 
  private GridBagLayout gbl;
  private GridBagConstraints gbc;

  private List nList;
  private Button OKbutton;
  private Button AddButton;
  private Button DeleteButton;
  private Button ChangeButton;

} // MacroSelect
