
/*  ----------------------------------------------------------------------
   sf Control Panel - A management software for sf Firewall
   Copyright (C) 1997 Roland E. Schmid <schmid@acm.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

package sfclasses;

import java.io.*;
import sfclasses.*;

/**
 * This class implements an input stream to read data from a firewall system
 * via the relay. It uses the I/O functions implemented in the Communicator
 * class. Filter streams can be used on top of this stream to enhance
 * functionality.
 * @version 1.0 11 Dec 1996
 * @author Roland E. Schmid
 */

public class RelayInputStream extends InputStream {

  /**
   * Initialize the stream and request the file from the firewall.
   * @param h IP address of the firewall
   * @param filename Name of the file to read including the full path name.
   * The file must be in the configuration directory <tt>/etc/firewall.d</tt>
   * or in the log directory <tt>/var/log</tt>.
   */
  public RelayInputStream(byte[] h, String filename) {
    host = h;
    EOF = !Communicator.requestFile(host, filename);
  }

  /**
   * Read one byte from the input stream
   * @return read byte or -1 on error
   */
  public int read() throws IOException {
    if (EOF) 
      return -1;
    if (bufptr < buffer.length) {
      bufptr++;
      return buffer[bufptr-1];
    }
    try {
      Communicator.readHeader();
      if (Communicator.readLength == 0) {
        EOF = true;
        return -1;
      }
      buffer = new byte[Communicator.readLength];
      Communicator.in.readFully(buffer, 0, Communicator.readLength);
      bufptr = 1;
      return buffer[0];
    }
    catch (Exception e) {
      throw new IOException();
    }
  }

  /**
   * Close the input stream.
   */
  public void close() throws IOException {
    while (!EOF) {
      try {
        Communicator.readHeader();
      }
      catch (Exception e) {
        throw new IOException();
      }
      if (Communicator.readLength == 0)
        EOF = true;
      else {
        buffer = new byte[Communicator.readLength];
        Communicator.in.readFully(buffer, 0, Communicator.readLength);
      }
    }
  }

  /**
   * Return number of bytes available for reading. If the buffer is
   * empty but the end of the file is not reached, 1 is returned.
   */
  public int available() {
    if (EOF)
      return 0;
    if (buffer.length > bufptr)
      return buffer.length - bufptr;
    return 1;
  } 

  /**
   * Check if the end of the file has been reached.
   * @return true if EOF, false otherwise
   */
  public boolean isEOF() {
    return EOF;
  }

  private byte[] host;
  private boolean EOF = true;
  private byte[] buffer = new byte[0];
  private int bufptr = 0;

}

