/*  ----------------------------------------------------------------------
   sf Firewall Software -- a TCP/IP packet filter for Linux
   Copyright (C) 1996 Robert Muchsel and Roland Schmid

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to 
   firewall-bugs@switch.ch.
   ----------------------------------------------------------------------  */

/*
 * $Id: sf_filter.h,v 1.20 1995/08/05 11:37:46 robby Rel $
 *
 */

#ifndef _SF_FILTER_H
#define _SF_FILTER_H

#include <linux/firewall.h>
#include <linux/timer.h>
#include "sf_global.h"

/*
 *    Format of the data structure which is passed to the 
 *      The size of the structure should not exceed a page size (4K).
 */

#define PROC_SIZE_IP		176

/* special rule identifiers for implicit rules */
#define RULE_SPOOF_RECV		-1	/* packet received on wrong interface */
#define RULE_SPOOF_XMIT		-2	/* packet sent via wrong interface */
#define RULE_OVERSIZED		-3	/* oversized packet received */
#define RULE_CONNECTION		-4	/* ftp data connection */
#define RULE_REJECT_FTP		-5	/* reject ftp data connection */

struct sf_proc_entry {
  int action;			/* action code (= return code)  */
  union {
    long num;			/* rule number                  */
    struct sf_fw *ptr;		/* pointer to rule              */
  } rule;
  char devname[8];		/* device address               */
  char ip_packet[PROC_SIZE_IP];	/* ip packet (first N bytes)    */
};

#define NUM_PROC_ENTRIES	21

struct sf_proc {
  unsigned num_entries;		/* number of valid records      */
  unsigned long lost_entries;	/* number of lost packets       */
  struct sf_proc_entry data[NUM_PROC_ENTRIES];
};

/* TCP circuits */

/* The client is the host that initiates the connection. */
#define SF_TCP_ACCEPT_SYN	0x00	/* accept client SYN   */

#define SF_TCP_CLIENT_SYN	0x01	/* client has sent SYN */
#define SF_TCP_SERVER_SYN	0x02	/* server has sent SYN ACK */
#define SF_TCP_SYN_ACK    	0x03	/* both of the above */

#define SF_TCP_CLIENT_ACK	0x04    /* client ACK of SYN ACK */
#define SF_TCP_ESTABLISHED3     0x07    /* connection is up */

#define SF_TCP_FTP		0x08    /* is ftp control session */
#define SF_TCP_ESTABLISHEDFTP   0x0F	/* ftp control connection is up */

#define SF_TCP_CLIENT_FIN	0x10	/* client has sent FIN */
#define SF_TCP_SERVER_FIN	0x20	/* server has sent FIN */
#define SF_TCP_TERMINATED	0x37	/* both of the above */

struct sf_tcp_connection {
  __u32 id;
  struct sf_tcp_connection *prev, *next;
  struct sf_tcp_connection *before, *after;
  __u32 client_addr, server_addr;
  __u16 client_port, server_port;
  __u16 client_win, server_win; /* window sizes */
  __u8 client_scale, server_scale; /* window scale factors */
  __u32 client_seq, server_seq; /* sequence numbers */
  int state;			/* state of tcp connection */
  struct timer_list timer;
  __u32 established;		/* "jiffies" connection was established */
  __u32 lastused;		/* "jiffies" connection was last used */
  struct sf_fw *rule;		/* accepted/created by rule */
  struct device *rif;		/* device of initial SYN */
  int timer_set:1,		/* one if a timer is set */
      allpsource:1,		/* one if all source ports match */
      log_open:1,		/* log beginning of connection */
      log_dxfer:1,		/* log data transfers (i.e. FTP) */
      ftp_allow_active:1,	/* active mode FTP is allowed */
      ftp_allow_passive:1,	/* passive mode FTP is allowed */
      check_sequence:1;		/* check TCP sequence numbers */
};

#define SF_TCP_HASH_SIZE	499	/* must be a prime number */
#define SF_TCP_KEY(a)  ( (a) % SF_TCP_HASH_SIZE )

#define FTP_CONTROL_PORT	21
#define FTP_DATA_PORT		20

/*
 *    Variables for interaction with the driver
 */

/* filter use count */
extern int sf_fw_enabled;

/* packet filtering routine */
extern int sf_check_packet(struct iphdr *, struct device *, int);

/* configuration update */
extern int sf_write_config(const char *, int);

/* temporary storage for packets to be sent to the daemon */
extern struct sf_proc sf_log_data;

/* wait queue for synchronization between kernel filter module and daemon */
extern struct wait_queue *sf_log_entry;

/* delete all timers, must be called before unloading the module */
extern void sf_del_all_timers(void);

/* initialize all data structures */
extern int sf_init(void);

/* read actual configuration */
extern struct sf_fw *prio_rules;
extern struct sf_fw *rules;
extern struct sf_tcp_connection *conns;
extern struct sf_fw *rule_first_next;
extern unsigned long conn_id;

extern inline long sf_rule_first(void) 
{
  /* return pointer to first rule */

  rule_first_next = prio_rules;
  return (long) rule_first_next;
}

extern inline long sf_rule_next(void)
{
/* return pointer to next rule */
  if (rule_first_next != NULL)
    rule_first_next = rule_first_next->fw_next;

  return (long) rule_first_next;
}

extern inline long sf_conn_first(void)
{
  /* return pointer to first active TCP connection */
  if (conns != NULL)
    conn_id = conns->id;
  else
    conn_id = 0;
#ifdef SF_DEBUG_LLIST
  printk("sf_conn_first: @  = %08x\n", (unsigned int) conns);
  if (conn_first_next != NULL) {
    printk("sf_conn_first: id = %08x\n", (unsigned int) conns->id);
    printk("sf_conn_first: <- = %08x\n", (unsigned int) conns->before);
    printk("sf_conn_first: -> = %08x\n", (unsigned int) conns->after);
  }
#endif

  return (long) conns;
}

extern inline long sf_conn_next(void)
{
  /* get pointer to next active TCP connection */
  /* we don't save a pointer to the list, because the list could change */
  /* between calls (i.e. an entry expires) */
  struct sf_tcp_connection *sftcp;

  sftcp = conns;
  while (sftcp != NULL && sftcp->id <= conn_id)
    sftcp = sftcp->after;
  if (sftcp != NULL)
    conn_id = sftcp->id;
  else
    conn_id = 0;

#ifdef SF_DEBUG_LLIST
  printk("sf_conn_next: @  = %08x\n", (unsigned int) sftcp);
  if (sftcp != NULL) {
    printk("sf_conn_next: id = %08x\n", (unsigned int) sftcp->id);
    printk("sf_conn_next: <- = %08x\n", (unsigned int) sftcp->before);
    printk("sf_conn_next: -> = %08x\n", (unsigned int) sftcp->after);
  }
#endif

  return (long) sftcp;
}

extern struct firewall_ops sf_fw_ops;
extern struct firewall_ops sf_blocking_fw_ops;

#endif				/* _SF_FILTER_H */
